import copy

from satlas import tqdm
from satlas.stats import fitting
import h5py
import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import uncertainties as u
from scipy import optimize
from scipy.stats import chi2

inv_color_list = ['#7acfff', '#fff466', '#00c48f', '#ff8626', '#ff9cd3', '#0093e6']
color_list = [c for c in reversed(inv_color_list)]
cmap = mpl.colors.ListedColormap(color_list)
cmap.set_over(color_list[-1])
cmap.set_under(color_list[0])
invcmap = mpl.colors.ListedColormap(inv_color_list)
invcmap.set_over(inv_color_list[-1])
invcmap.set_under(inv_color_list[0])

__all__ = ['generate_correlation_map', 'generate_correlation_plot', 'generate_walk_plot']

def _make_axes_grid(no_variables, padding=0, cbar_size=0.5, axis_padding=0.5, cbar=True):
    """Makes a triangular grid of axes, with a colorbar axis next to it.

    Parameters
    ----------
    no_variables: int
        Number of variables for which to generate a figure.
    padding: float
        Padding around the figure (in cm).
    cbar_size: float
        Width of the colorbar (in cm).
    axis_padding: float
        Padding between axes (in cm).

    Returns
    -------
    fig, axes, cbar: tuple
        Tuple containing the figure, a 2D-array of axes and the colorbar axis."""

    # Convert to inches.
    padding, cbar_size, axis_padding = (padding * 0.393700787,
                                        cbar_size * 0.393700787,
                                        axis_padding * 0.393700787)
    if not cbar:
        cbar_size = 0

    # Generate the figure, convert padding to percentages.
    fig = plt.figure()
    padding = 1

    axis_size_left = (fig.get_figwidth()-padding - 0*(no_variables + 1) * padding) / no_variables
    axis_size_up = (fig.get_figheight()-padding - 0*(no_variables + 1) * padding) / no_variables

    cbar_size = cbar_size / fig.get_figwidth()
    left_padding = padding * 0.5 / fig.get_figwidth()
    left_axis_padding = axis_padding / fig.get_figwidth()
    up_padding = padding * 0.5 / fig.get_figheight()
    up_axis_padding = 0*axis_padding / fig.get_figheight()
    axis_size_left = axis_size_left / fig.get_figwidth()
    axis_size_up = axis_size_up / fig.get_figheight()

    # Pre-allocate a 2D-array to hold the axes.
    axes = np.array([[None for _ in range(no_variables)] for _ in range(no_variables)],
                    dtype='object')

    for i, I in zip(range(no_variables), reversed(range(no_variables))):
        for j in reversed(range(no_variables)):
            # Only create axes on the lower triangle.
            if I + j < no_variables:
                # Share the x-axis with the plot on the diagonal,
                # directly above the plot.
                sharex = axes[j, j] if i != j else None
                # Share the y-axis among the 2D maps along one row,
                # but not the plot on the diagonal!
                sharey = axes[i, i-1] if (i != j and i-1 != j) else None
                # Determine the place and size of the axes
                left_edge = j * axis_size_left + left_padding
                bottom_edge = I * axis_size_up + up_padding
                if j > 0:
                    left_edge += j * left_axis_padding
                if I > 0:
                    bottom_edge += I * up_axis_padding

                a = plt.axes([left_edge, bottom_edge, axis_size_left, axis_size_up],
                             sharex=sharex, sharey=sharey)
                plt.setp(a.xaxis.get_majorticklabels(), rotation=45)
                plt.setp(a.yaxis.get_majorticklabels(), rotation=45)
            else:
                a = None
            if i == j:
                a.yaxis.tick_right()
                a.yaxis.set_label_position('right')
            axes[i, j] = a

    axes = np.array(axes)
    for a in axes[:-1, :].flatten():
        if a is not None:
            plt.setp(a.get_xticklabels(), visible=False)
    for a in axes[:, 1:].flatten():
        if a is not None:
            plt.setp(a.get_yticklabels(), visible=False)
    left_edge = no_variables*(axis_size_left+left_axis_padding)+left_padding
    bottom_edge = up_padding
    width = cbar_size

    height = axis_size_up * len(axes) + up_padding * (len(axes) - 1)

    cbar_width = axis_size_left * 0.1
    if cbar:
        cbar = plt.axes([1-cbar_width-padding*0.5/fig.get_figwidth(), padding*0.5/fig.get_figheight()+axis_size_up*1.5, cbar_width, axis_size_up*(no_variables-1)-axis_size_up*0.5])
        plt.setp(cbar.get_xticklabels(), visible=False)
        plt.setp(cbar.get_yticklabels(), visible=False)
    else:
        cbar = None
    return fig, axes, cbar

def generate_correlation_map(f, x_data, y_data, method='chisquare_spectroscopic', filter=None, resolution_diag=20, resolution_map=15, fit_args=tuple(), fit_kws={}, distance=5, npar=1):
    """Generates a correlation map for either the chisquare or the MLE method.
    On the diagonal, the chisquare or loglikelihood is drawn as a function of one fixed parameter.
    Refitting to the data each time gives the points on the line. A dashed line is drawn on these
    plots, with the intersection with the plots giving the correct confidence interval for the
    parameter. In solid lines, the interval estimated by the fitting routine is drawn.
    On the offdiagonal, two parameters are fixed and the model is again fitted to the data.
    The change in chisquare/loglikelihood is mapped to 1, 2 and 3 sigma contourmaps.

    Parameters
    ----------
    f: :class:`.BaseModel`
        Instance of the model for which the contour map has to be generated.
    x_data: array_like or list of array_likes
        Data on the x-axis for the fit. Must be appropriate input for *f*.
    y_data: array_like or list of array_likes
        Data on the y-axis for the fit. Must be appropriate input for *f*.

    Other parameters
    ----------------
    method: {'chisquare', 'chisquare_spectroscopic', mle'}
        Chooses between generating the map for the chisquare routine or for
        the likelihood routine.
    filter: list of strings
        Only the parameters matching the names given in this list will be used
        to generate the maps.
    resolution_diag: int
        Number of points for the line plot on each diagonal.
    resolution_map: int
        Number of points along each dimension for the meshgrids.
    fit_kws: dictionary
        Dictionary of keywords to pass on to the fitting routine.
    npar: int
        Number of parameters for which simultaneous predictions need to be made.
        Influences the uncertainty estimates from the parabola."""

    # Save the original goodness-of-fit and parameters for later use
    mapping = {'chisquare_spectroscopic': (fitting.chisquare_spectroscopic_fit, 'chisqr_chi'),
               'chisquare': (fitting.chisquare_fit, 'chisqr_chi'),
               'mle': (fitting.likelihood_fit, 'likelihood_mle')}
    func, attr = mapping.pop(method.lower(), (fitting.chisquare_spectroscopic_fit, 'chisqr_chi'))
    title = '{}\n${}_{{-{}}}^{{+{}}}$'
    title_e = '{}\n$({}_{{-{}}}^{{+{}}})e{}$'
    fit_kws['verbose'] = False
    fit_kws['hessian'] = False

    to_save = {'mle': ('fit_mle', 'result_mle')}
    to_save = to_save.pop(method.lower(), ('chisq_res_par', 'ndof_chi', 'redchi_chi'))
    saved = [copy.deepcopy(getattr(f, attr)) for attr in to_save]

    # func(f, x_data, y_data, *fit_args, **fit_kws)

    orig_value = getattr(f, attr)
    orig_params = copy.deepcopy(f.params)
    state = fitting._get_state(f, method=method.lower())

    ranges = {}

    chifunc = lambda x: chi2.cdf(x, npar) - 0.682689492 # Calculate 1 sigma boundary
    boundary = optimize.root(chifunc, npar).x[0] * 0.5 if method.lower() == 'mle' else optimize.root(chifunc, npar).x[0]
    # Select all variable parameters, generate the figure
    param_names = []
    no_params = 0
    for p in orig_params:
        if orig_params[p].vary and (filter is None or any([f in p for f in filter])):
            no_params += 1
            param_names.append(p)
    fig, axes, cbar = _make_axes_grid(no_params, axis_padding=0, cbar=no_params > 1)

    # Make the plots on the diagonal: plot the chisquare/likelihood
    # for the best fitting values while setting one parameter to
    # a fixed value.
    saved_params = copy.deepcopy(f.params)
    function_kws = {'method': method.lower(), 'func_args': fit_args, 'func_kwargs': fit_kws}
    function_kws['orig_stat'] = orig_value
    for i in range(no_params):
        params = copy.deepcopy(saved_params)
        ranges[param_names[i]] = {}

        # Set the y-ticklabels.
        ax = axes[i, i]
        ax.set_title(param_names[i])
        if i == no_params-1:
            if method.lower().startswith('chisquare'):
                ax.set_ylabel(r'$\Delta\chi^2$')
            else:
                ax.set_ylabel(r'$\Delta\mathcal{L}$')

        # Select starting point to determine error widths.
        value = orig_params[param_names[i]].value
        stderr = orig_params[param_names[i]].stderr
        print(stderr)
        stderr = stderr if stderr is not None else 0.01 * np.abs(value)
        stderr = stderr if stderr != 0 else 0.01 * np.abs(value)
        result_left, success_left = fitting._find_boundary(-stderr, param_names[i], boundary, f, x_data, y_data, function_kwargs=function_kws)
        result_right, success_right = fitting._find_boundary(stderr, param_names[i], boundary, f, x_data, y_data, function_kwargs=function_kws)
        success = success_left * success_right
        ranges[param_names[i]]['left'] = result_left
        ranges[param_names[i]]['right'] = result_right

        if not success:
            print("Warning: boundary calculation did not fully succeed for " + param_names[i])
        right = np.abs(ranges[param_names[i]]['right'] - value)
        left = np.abs(ranges[param_names[i]]['left'] - value)
        params[param_names[i]].vary = False

        left_val, right_val = max(value - distance * left, orig_params[param_names[i]].min), min(value + distance * right, orig_params[param_names[i]].max)
        ranges[param_names[i]]['right_val'] = right_val
        ranges[param_names[i]]['left_val'] = left_val
        value_range = np.linspace(left_val, right_val, resolution_diag)
        value_range = np.sort(np.append(value_range, np.array([value - left, value + right, value])))
        chisquare = np.zeros(len(value_range))
        # Calculate the new value, and store it in the array. Update the progressbar.
        with tqdm.tqdm(value_range, desc=param_names[i], leave=True) as pbar:
            for j, v in enumerate(value_range):
                chisquare[j] = fitting.calculate_updated_statistic(v, param_names[i], f, x_data, y_data, **function_kws)
                fitting._set_state(f, state, method=method.lower())
                pbar.update(1)
        # Plot the result
        ax.plot(value_range, chisquare, color='k')

        c = '#0093e6'
        # Indicate the used interval.
        ax.axvline(value + right, ls="dashed", color=c)
        ax.axvline(value - left, ls="dashed", color=c)
        ax.axvline(value, ls="dashed", color=c)
        ax.axhline(boundary, color=c)
        up = '{:.2ug}'.format(u.ufloat(value, right))
        down = '{:.2ug}'.format(u.ufloat(value, left))
        val = up.split('+')[0].split('(')[-1]
        r = up.split('-')[1].split(')')[0]
        l = down.split('-')[1].split(')')[0]
        if 'e' in up or 'e' in down:
            ex = up.split('e')[-1]
            ax.set_title(title_e.format(param_names[i], val, l, r, ex))
        else:
            ax.set_title(title.format(param_names[i], val, l, r))
        # Restore the parameters.
        fitting._set_state(f, state, method=method.lower())

    for i, j in zip(*np.tril_indices_from(axes, -1)):
        params = copy.deepcopy(orig_params)
        ax = axes[i, j]
        x_name = param_names[j]
        y_name = param_names[i]
        if j == 0:
            ax.set_ylabel(y_name)
        if i == no_params - 1:
            ax.set_xlabel(x_name)
        right = ranges[x_name]['right_val']
        left = ranges[x_name]['left_val']
        x_range = np.append(np.linspace(left, right, resolution_map), orig_params[x_name].value)
        x_range = np.sort(x_range)

        right = ranges[y_name]['right_val']
        left = ranges[y_name]['left_val']
        y_range = np.append(np.linspace(left, right, resolution_map), orig_params[y_name].value)
        y_range = np.sort(y_range)

        X, Y = np.meshgrid(x_range, y_range)
        Z = np.zeros(X.shape)
        i_indices, j_indices = np.indices(Z.shape)
        with tqdm.tqdm(i_indices.flatten(), desc=param_names[j]+' ' + param_names[i], leave=True) as pbar:
            for k, l in zip(i_indices.flatten(), j_indices.flatten()):
                x = X[k, l]
                y = Y[k, l]
                print(x, y, f.params['Background0'].value)
                Z[k, l] = fitting.calculate_updated_statistic([x, y], [x_name, y_name], f, x_data, y_data, **function_kws)
                fitting._set_state(f, state, method=method.lower())

                pbar.update(1)
        Z = -Z
        npar = 1
        bounds = []
        for bound in [0.997300204, 0.954499736, 0.682689492]:
            chifunc = lambda x: chi2.cdf(x, npar) - bound # Calculate 1 sigma boundary
            bounds.append(-optimize.root(chifunc, npar).x[0])
        # bounds = sorted([-number*number for number in np.arange(1, 9, .1)])
        bounds.append(1)
        if method.lower() == 'mle':
            bounds = [b * 0.5 for b in bounds]
        norm = mpl.colors.BoundaryNorm(bounds, invcmap.N)
        contourset = ax.contourf(X, Y, Z, bounds, cmap=invcmap, norm=norm)
        f.params = copy.deepcopy(orig_params)
    if method.lower() == 'mle':
        f.fit_mle = copy.deepcopy(orig_params)
    else:
        f.chisq_res_par
    try:
        cbar = plt.colorbar(contourset, cax=cbar, orientation='vertical')
        cbar.ax.yaxis.set_ticks([0, 1/6, 0.5, 5/6])
        cbar.ax.set_yticklabels(['', r'3$\sigma$', r'2$\sigma$', r'1$\sigma$'])
    except:
        pass
    setattr(f, attr, orig_value)
    for attr, value in zip(to_save, saved):
        setattr(f, attr, copy.deepcopy(value))
    for a in axes.flatten():
        if a is not None:
            for label in a.get_xticklabels()[::2]:
                label.set_visible(False)
            for label in a.get_yticklabels()[::2]:
                label.set_visible(False)
    return fig, axes, cbar

def generate_correlation_plot(filename, filter=None, bins=None, selection=(0, 100)):
    """Given the random walk data, creates a triangle plot: distribution of
    a single parameter on the diagonal axes, 2D contour plots with 1, 2 and
    3 sigma contours on the off-diagonal. The 1-sigma limits based on the
    percentile method are also indicated, as well as added to the title.

    Parameters
    ----------
    filename: string
        Filename for the h5 file containing the data from the walk.
    filter: list of str, optional
        If supplied, only this list of columns is used for the plot.
    bins: int or list of int, optional
        If supplied, use this number of bins for the plotting.

    Returns
    -------
    figure
        Returns the MatPlotLib figure created."""

    with h5py.File(filename, 'r') as store:
        columns = store['data'].attrs['format']
        columns = [f.decode('utf-8') for f in columns]
        if filter is not None:
            filter = [c for f in filter for c in columns if f in c]
        else:
            filter = columns
        with tqdm.tqdm(total=len(filter)+(len(filter)**2-len(filter))/2, leave=True) as pbar:
            fig, axes, cbar = _make_axes_grid(len(filter), axis_padding=0)

            metadata = {}
            if not isinstance(bins, list):
                bins = [bins for _ in filter]
            dataset_length = store['data'].shape[0]
            first, last = int(np.floor(dataset_length/100*selection[0])), int(np.ceil(dataset_length/100*selection[1]))
            for i, val in enumerate(filter):
                pbar.set_description(val)
                ax = axes[i, i]
                bin_index = i
                i = columns.index(val)
                x = store['data'][first:last, i]
                if bins[bin_index] is None:
                    width = 3.5*np.std(x)/x.size**(1/3) #Scott's rule for binwidth
                    bins[bin_index] = np.arange(x.min(), x.max()+width, width)
                try:
                    n, b, p, = ax.hist(x, int(bins[bin_index]), histtype='step', color='k')
                except TypeError:
                    bins[bin_index] = 50
                    n, b, p, = ax.hist(x, int(bins[bin_index]), histtype='step', color='k')
                center = n.argmax()
                q50 = (b[center] + b[center+1])/2
                # q16 = np.percentile(x[x<q50], 84)
                # q84 = np.percentile(x[x>q50], 16)
                metadata[val] = {'bins': bins[bin_index], 'min': x.min(), 'max': x.max()}

                q = [16.0, 84.0]
                q16, q84 = np.percentile(x, q)

                title = '{}\n${}_{{-{}}}^{{+{}}}$'
                title_e = '{}\n$({}_{{-{}}}^{{+{}}})e{}$'
                up = '{:.2ug}'.format(u.ufloat(q50, np.abs(q84-q50)))
                down = '{:.2ug}'.format(u.ufloat(q50, np.abs(q50-q16)))
                param_val = up.split('+')[0].split('(')[-1]
                r = up.split('+/-')[1].split(')')[0]
                l = down.split('+/-')[1].split(')')[0]
                if 'e' in up or 'e' in down:
                    ex = up.split('e')[-1]
                    ax.set_title(title_e.format(val, param_val, l, r, ex))
                else:
                    ax.set_title(title.format(val, param_val, l, r))

                qvalues = [q16, q50, q84]
                c = '#0093e6'
                for q in qvalues:
                    ax.axvline(q, ls="dashed", color=c)
                ax.set_yticks([])
                ax.set_yticklabels([])
                pbar.update(1)

            for i, j in zip(*np.tril_indices_from(axes, -1)):
                x_name = filter[j]
                y_name = filter[i]
                pbar.set_description(x_name + ' ' + y_name)
                ax = axes[i, j]
                if j == 0:
                    ax.set_ylabel(filter[i])
                if i == len(filter) - 1:
                    ax.set_xlabel(filter[j])
                j = columns.index(x_name)
                i = columns.index(y_name)
                x = store['data'][first:last, j]
                y = store['data'][first:last, i]
                x_min, x_max, x_bins = metadata[x_name]['min'], metadata[x_name]['max'], metadata[x_name]['bins']
                y_min, y_max, y_bins = metadata[y_name]['min'], metadata[y_name]['max'], metadata[y_name]['bins']
                X = np.linspace(x_min, x_max, x_bins + 1)
                Y = np.linspace(y_min, y_max, y_bins + 1)
                H, X, Y = np.histogram2d(x.flatten(), y.flatten(), bins=(X, Y),
                                         weights=None)
                X1, Y1 = 0.5 * (X[1:] + X[:-1]), 0.5 * (Y[1:] + Y[:-1])
                X, Y = X[:-1], Y[:-1]
                H = (H - H.min()) / (H.max() - H.min())

                Hflat = H.flatten()
                inds = np.argsort(Hflat)[::-1]
                Hflat = Hflat[inds]
                sm = np.cumsum(Hflat)
                sm /= sm[-1]
                levels = 1.0 - np.exp(-0.5 * np.arange(1, 3.1, 1) ** 2)
                V = np.empty(len(levels))
                for i, v0 in enumerate(levels):
                    try:
                        V[i] = Hflat[sm <= v0][-1]
                    except:
                        V[i] = Hflat[0]

                bounds = np.unique(np.concatenate([[H.max()], V])[::-1])
                norm = mpl.colors.BoundaryNorm(bounds, invcmap.N)

                contourset = ax.contourf(X1, Y1, H.T, bounds, cmap=invcmap, norm=norm)
                pbar.update(1)
            try:
                cbar = plt.colorbar(contourset, cax=cbar, orientation='vertical')
                cbar.ax.yaxis.set_ticks([0, 1/6, 0.5, 5/6])
                cbar.ax.set_yticklabels(['', r'3$\sigma$', r'2$\sigma$', r'1$\sigma$'])
            except:
                cbar = None
    return fig, axes, cbar

def generate_walk_plot(filename, filter=None, selection=(0, 100), walkers=20):
    """Given the random walk data, the random walk for the selected parameters
    is plotted.

    Parameters
    ----------
    filename: string
        Filename for the h5 file containing the data from the walk.
    filter: list of str, optional
        If supplied, only this list of parameters is used for the plot.

    Returns
    -------
    figure
        Returns the MatPlotLib figure created."""

    with h5py.File(filename, 'r') as store:
        columns = store['data'].attrs['format']
        columns = [f.decode('utf-8') for f in columns]
        if filter is not None:
            filter = [c for f in filter for c in columns if f in c]
        else:
            filter = columns
        with tqdm.tqdm(total=len(filter)+(len(filter)**2-len(filter))/2, leave=True) as pbar:
            fig, axes = plt.subplots(len(filter), 1, sharex=True)

            dataset_length = store['data'].shape[0]
            first, last = int(np.floor(dataset_length/100*selection[0])), int(np.ceil(dataset_length/100*selection[1]))
            for i, (val, ax) in enumerate(zip(filter, axes)):
                pbar.set_description(val)
                i = columns.index(val)
                x = store['data'][first:last, i]
                new_x = np.reshape(x, (-1, walkers))
                q50 = np.percentile(x, [50.0])
                ax.plot(range(new_x.shape[0]), new_x, alpha=0.3, color='gray')
                ax.set_ylabel(val)
                ax.axhline(q50, color='k')
                pbar.update(1)
            ax.set_xlabel('Step')
        pbar.close()
    return fig, axes



from ._tqdm import tqdm
from ._tqdm import trange
from ._tqdm_gui import tqdm_gui
from ._tqdm_gui import tgrange
from ._tqdm_pandas import tqdm_pandas
from ._version import __version__  # NOQA

__all__ = ['tqdm', 'tqdm_gui', 'trange', 'tgrange', 'tqdm_pandas',
           '__version__']



"""
Collection of style setting functions using both style sheets in the subfolder 'styles' and user-supplied adjustments.

.. moduleauthor:: Wouter Gins <wouter.gins@kuleuven.be
"""
import matplotlib as mpl
import contextlib
import os
import glob

STYLE_PATH = os.path.dirname(os.path.realpath(__file__)) + '\\styles\\'
STYLES = [s.split('\\')[-1].split('.')[0] for s in glob.glob(STYLE_PATH + '*.mplstyle')]
STYLE_MAPPING = {s: os.path.dirname(os.path.realpath(__file__)) + '\\styles\\' + s + '.mplstyle' for s in STYLES}

__all__ = ['set', 'get_available_styles', 'context', 'set_font']

def set(style='standard'):
    """Sets the current style to a SATLAS style. For
    a list of available styles, use :func:`.get_available_styles()`.

    Parameters
    ----------
    style: str or list
        Style sheets specification. Valid options are:

        +------+----------------------------------------------------+
        | str  | The name of a SATLAS style.                        |
        +------+----------------------------------------------------+
        | list | A list of style specifiers (str) applied from first|
        |      | to last in the list.                               |
        +------+----------------------------------------------------+"""
    if mpl.cbook.is_string_like(style):
        style = [style]
    style = [STYLE_MAPPING[s] for s in style]
    mpl.style.use(style)

def get_available_styles():
    """Returns the available stylesheets in the subfolder 'styles'."""
    return STYLES

@contextlib.contextmanager
def context(style, after_reset=False):
    """Context manager for using style settings temporarily.

    Parameters
    ----------
    style: str or list
        Style sheets specification. Valid options are:

        +------+----------------------------------------------------+
        | str  | The name of a SATLAS style.                        |
        +------+----------------------------------------------------+
        | list | A list of style specifiers (str) applied from first|
        |      | to last in the list.                               |
        +------+----------------------------------------------------+

    after_reset : bool
        If True, apply style after resetting settings to their defaults;
        otherwise, apply style on top of the current settings.
    """
    if mpl.cbook.is_string_like(style):
        style = [style]
    style = [STYLE_MAPPING[s] for s in style]
    mpl.style.context(style, after_rest=after_reset)

def set_font(font='Palatino Linotype', family='serif'):
    """Sets the font to the chosen family and the family options
    to the supplied fonts:

    Parameters
    ----------
    font: str or list
        Name or list of names of fonts.
    family: str
        Family name of the fonts that will be set and used."""
    d = {'font.family': family, 'font.' + family: font}
    mpl.rcParams.update(d)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
A vanilla Metropolis-Hastings sampler

"""

from __future__ import (division, print_function, absolute_import,
                        unicode_literals)

__all__ = ["MHSampler"]

import numpy as np

from . import autocorr
from .sampler import Sampler


# === MHSampler ===
class MHSampler(Sampler):
    """
    The most basic possible Metropolis-Hastings style MCMC sampler

    :param cov:
        The covariance matrix to use for the proposal distribution.

    :param dim:
        Number of dimensions in the parameter space.

    :param lnpostfn:
        A function that takes a vector in the parameter space as input and
        returns the natural logarithm of the posterior probability for that
        position.

    :param args: (optional)
        A list of extra positional arguments for ``lnpostfn``. ``lnpostfn``
        will be called with the sequence ``lnpostfn(p, *args, **kwargs)``.

    :param kwargs: (optional)
        A list of extra keyword arguments for ``lnpostfn``. ``lnpostfn``
        will be called with the sequence ``lnpostfn(p, *args, **kwargs)``.

    """
    def __init__(self, cov, *args, **kwargs):
        super(MHSampler, self).__init__(*args, **kwargs)
        self.cov = cov

    def reset(self):
        super(MHSampler, self).reset()
        self._chain = np.empty((0, self.dim))
        self._lnprob = np.empty(0)

    def sample(self, p0, lnprob=None, randomstate=None, thin=1,
               storechain=True, iterations=1):
        """
        Advances the chain ``iterations`` steps as an iterator

        :param p0:
            The initial position vector.

        :param lnprob0: (optional)
            The log posterior probability at position ``p0``. If ``lnprob``
            is not provided, the initial value is calculated.

        :param rstate0: (optional)
            The state of the random number generator. See the
            :func:`random_state` property for details.

        :param iterations: (optional)
            The number of steps to run.

        :param thin: (optional)
            If you only want to store and yield every ``thin`` samples in the
            chain, set thin to an integer greater than 1.

        :param storechain: (optional)
            By default, the sampler stores (in memory) the positions and
            log-probabilities of the samples in the chain. If you are
            using another method to store the samples to a file or if you
            don't need to analyse the samples after the fact (for burn-in
            for example) set ``storechain`` to ``False``.

        At each iteration, this generator yields:

        * ``pos`` - The current positions of the chain in the parameter
          space.

        * ``lnprob`` - The value of the log posterior at ``pos`` .

        * ``rstate`` - The current state of the random number generator.

        """

        self.random_state = randomstate

        p = np.array(p0)
        if lnprob is None:
            lnprob = self.get_lnprob(p)

        # Resize the chain in advance.
        if storechain:
            N = int(iterations / thin)
            self._chain = np.concatenate((self._chain,
                                          np.zeros((N, self.dim))), axis=0)
            self._lnprob = np.append(self._lnprob, np.zeros(N))

        i0 = self.iterations
        # Use range instead of xrange for python 3 compatability
        for i in range(int(iterations)):
            self.iterations += 1

            # Calculate the proposal distribution.
            q = self._random.multivariate_normal(p, self.cov)
            newlnprob = self.get_lnprob(q)
            diff = newlnprob - lnprob

            # M-H acceptance ratio
            if diff < 0:
                diff = np.exp(diff) - self._random.rand()

            if diff > 0:
                p = q
                lnprob = newlnprob
                self.naccepted += 1

            if storechain and i % thin == 0:
                ind = i0 + int(i / thin)
                self._chain[ind, :] = p
                self._lnprob[ind] = lnprob

            # Heavy duty iterator action going on right here...
            yield p, lnprob, self.random_state

    @property
    def acor(self):
        """
        An estimate of the autocorrelation time for each parameter (length:
        ``dim``).

        """
        return self.get_autocorr_time()

    def get_autocorr_time(self, window=50):
        """
        Compute an estimate of the autocorrelation time for each parameter
        (length: ``dim``).

        :param window: (optional)
            The size of the windowing function. This is equivalent to the
            maximum number of lags to use. (default: 50)

        """
        return autocorr.integrated_time(self.chain, axis=0, window=window)



import csv
import pandas as pd
from . import phys_calc
import matplotlib.pyplot as plt
import numpy as np
import os
import brokenaxes
import ast

class HFS_data:
    def __init__(self, x_axis_name: str = 'Wavenumber', file_path: str | None = None):
        self.x_axis_name = x_axis_name
        self.df = pd.DataFrame(columns=['Timestamp', 'BunchNo', 'Channel', 'TOF', self.x_axis_name])
    
    def read_csv(self, file_path: str):
        """
        Parameters:
        - file_path: Path to the CSV file
        """
        timestamps = []
        wavenumbers = []
        signals = []
        with open(file_path, mode='r', encoding='utf-8') as file:
            csv_reader = csv.reader(file)
            next(csv_reader)  # Skip header
            for row in csv_reader:
                timestamps.append(float(row[0]))
                wavenumbers.append(float(row[1]))
                signals.append(eval(row[2]))
        data = []
        for i in range(len(signals)):
            for signal in signals[i]:
                data.append([timestamps[i], *signal, wavenumbers[i]])
        
        if self.df.empty:
            self.df = pd.DataFrame(data, columns=self.df.columns)
        else:
            self.df = pd.concat([self.df, pd.DataFrame(data, columns=self.df.columns)], ignore_index=True)

        self.df.sort_values(by='Timestamp', inplace=True)

    def read_new_csv(self, file_path: str, x_axis_name: str = 'Voltage'):
        """
        This is a method to read the new CSV file format.

        Parameters:
        - file_path: Path to the CSV file
        """
        df_new = pd.read_csv(file_path)
        df_new.dropna(axis=1, how='all', inplace=True)
        if x_axis_name not in df_new.columns:
            print(f"Warning: {x_axis_name} column not found in the CSV file")
            return 0
        if x_axis_name != 'Voltage':
            df_new.rename(columns={x_axis_name: 'Wavenumber'}, inplace=True)
        df_new.rename(columns={'time': 'Timestamp'}, inplace=True)
        df_new['data'] = df_new['data'].apply(lambda x: ast.literal_eval(x))
        df_new = df_new.explode('data')
        df_new.dropna(subset=['data'], inplace=True)
        df_new_expanded = df_new['data'].apply(pd.Series)
        df_new_expanded.columns = ['BunchNo', 'Channel', 'TOF']
        df_new = pd.concat([df_new.drop('data', axis = 1), df_new_expanded], axis=1)
        
        if self.df.empty:
            self.df = df_new
        else:
            self.df = pd.concat([self.df, pd.DataFrame(df_new, columns=self.df.columns)], ignore_index=True)

    def read_folder(self, folder_path: str):
        """
        Only for old version. This will be removed after New DAQ is completed.

        Parameters:
        - folder_path: Path to the folder containing CSV files
        """
        txt_files = [f for f in os.listdir(folder_path) if f.endswith('.txt')]
        timestamps = []
        voltages = []
        signals = []
        for txt_file in txt_files:
            with open(os.path.join(folder_path, txt_file), mode='r', encoding='utf-8') as file:
                file_data = eval(file.readline())
                timestamps.append(float(file_data[0]))
                voltages.append(float(file_data[1]))
                signals.append(file_data[2])
        
        data = []
        for i in range(len(signals)):
            for signal in signals[i]:
                data.append([timestamps[i], *signal, voltages[i]])
        
        if self.df.empty:
            self.df = pd.DataFrame(data, columns=self.df.columns)
        else:
            self.df = pd.concat([self.df, pd.DataFrame(data, columns=self.df.columns)], ignore_index=True)

        self.df.sort_values(by='Timestamp', inplace=True)

    def save_csv(self, file_path: str):
        """
        Parameters:
        - file_path: Path to save the DataFrame
        """

        # self.df['BunchNo'] = self.df['BunchNo'].astype('int')
        # self.df['Channel'] = self.df['Channel'].astype('int')
        # self.df['TOF'] = self.df['TOF'].astype('int')
        # self.df.drop(columns=['BunchNo'], inplace=True)

        self.df.to_csv(file_path, index=False)

    def read_cali_csv(self, file_path: str, name: str = 'InitEnergy', omit_value = None, mean_window: int = 0, is_draw: bool = True):
        """
        Parameters:
        - file_path: Path to the calibration CSV file
        """
        calibration = pd.read_csv(file_path)
        calibration.columns = ['Timestamp', name]
        calibration.dropna(subset=[name], inplace=True)

        if omit_value is not None:
            calibration = calibration[calibration[name] != omit_value]

        if mean_window > 0:
            calibration[name] = calibration[name].rolling(window=mean_window, center=True).mean()
            calibration.dropna(subset=[name], inplace=True)

        if is_draw:
            plt.plot(calibration['Timestamp'], calibration[name], '.-')
            plt.xlabel('Timestamp (s)')
            plt.ylabel(name)
            plt.title(f'{name} Calibration')
            plt.show()

        # 合并calibration数据到self.df中，保留所有数据
        if name not in self.df.columns:
            self.df = pd.merge(self.df, calibration, on='Timestamp', how='outer')
        else:
            self.df = pd.merge(self.df, calibration, on=['Timestamp', name], how='outer')

    def read_cali_txt(self, file_path: str, name: str = 'InitEnergy', omit_value = None):
        """
        Only for the old version. This will be removed after New DAQ is completed.
        Parameters:
        - file_path: Path to the calibration text file
        """
        with open(file_path, mode='r', encoding='utf-8') as file:
            file_data = eval(file.readline())

        calibration = pd.DataFrame(file_data, columns = ['Timestamp', name])
        calibration.dropna(subset=[name], inplace=True)

        if omit_value is not None:
            calibration = calibration[calibration[name] != omit_value]

        # 合并calibration数据到self.df中，保留所有数据
        if name not in self.df.columns:
            self.df = pd.merge(self.df, calibration, on='Timestamp', how='outer')
        else:
            self.df = pd.merge(self.df, calibration, on=['Timestamp', name], how='outer')
    
    def fill_cali(self, col_name: str, method: str = 'ffill'):
        """
        Parameters:
        - col_name: Name of the column to fill
        - method: Method to fill the column, 'ffill', 'bfill' or 'interpolate'
        """
        if method == 'ffill':
            self.df[col_name] = self.df[col_name].ffill().bfill()
        elif method == 'bfill':
            self.df[col_name] = self.df[col_name].bfill().ffill()
        elif method == 'linear':
            self.df[col_name] = self.df[col_name].interpolate().ffill().bfill()
        elif method == 'nearest':
            mask = self.df[col_name].isna()
            if mask.any():
                filled = self.df[~mask]
                for idx in self.df[mask].index:
                    ts = self.df.at[idx, 'Timestamp']
                    nearest_idx = (filled['Timestamp'] - ts).abs().idxmin()
                    self.df.at[idx, col_name] = filled.at[nearest_idx, col_name]

    def dropna(self):
        self.df = self.df.dropna(subset=['TOF'])

    def voltage_cali(self, BOP_file_path: str | None = None, gain_factor: float = 0.9988, is_new: bool = False):
        
        self.df['InitEnergy'] = self.df['InitEnergy'] * gain_factor
        
        if BOP_file_path is not None:
            with open(BOP_file_path) as file:
                reader = csv.reader(file)
                volt_in = []
                volt_out = []
                next(reader)
                for row in reader:
                    if is_new:
                        volt_in.append(float(row[0]))
                        volt_out.append(float(row[1]))
                    else:
                        time = float(row[0])
                        volt_in.append(float(row[1]))
                        volt_out.append(float(row[2]))
            k, b = np.polyfit(volt_in, volt_out, 1)  # 1 表示线性拟合
            # est=sm.OLS(volt_out,sm.add_constant(volt_in)).fit()
            # b, k = est.params
            self.df['InitEnergy'] = self.df['InitEnergy'] - k * self.df['Voltage'] - b
            self.df.drop(columns=['Voltage'], inplace=True)

    def read_voltage_cali_csv(self, BOP_file_path: str, gain_factor: float = 0.9988):
        with open(BOP_file_path) as file:
            reader = csv.reader(file)
            volt_in = []
            volt_out = []
            next(reader)
            for row in reader:
                timestamp = float(row[0])
                volt_in.append(float(row[1]))
                volt_out.append(float(row[2]))
        k, b = np.polyfit(volt_in, volt_out, 1)
        new_row = {'Timestamp': timestamp, 'BopK': k, 'BopB': b, 'GainFactor': gain_factor}
        df_bop = pd.DataFrame([new_row])
        self.df = pd.concat([self.df, df_bop], ignore_index=True)
        self.df.sort_values(by='Timestamp', inplace=True)
        self.df.reset_index(drop=True, inplace=True)
    
    def voltage_cali_with_bop(self):
        self.df['InitEnergy'] = self.df['InitEnergy'] * self.df['GainFactor'] - self.df['BopK'] * self.df['Voltage'] - self.df['BopB']
        self.df.drop(columns=['Voltage', 'BopK', 'BopB', 'GainFactor'], inplace=True)

    def diode_cali(self, ref_freq: float):
        """
        This is a method to calibrate the wavenumber using the diode signal.

        Parameters:
        - ref_freq: Reference frequency for the calibration
        """
        self.df['Wavenumber'] = self.df['Diode'] - ref_freq + self.df['Wavenumber']
        self.df.drop(columns=['Diode'], inplace=True)

    def doppler_shift(self, mass: float):
        """
        This is a method to calculate the Doppler shift.

        Parameters:
        - mass: Mass of the atom
        """
        self.df['Wavenumber'] = phys_calc.dopplerfactor(mass, self.df['InitEnergy'] * 1000) * self.df['Wavenumber']
        self.df.drop(columns=['InitEnergy'], inplace=True)

    def doppler_shift_for_2_photons(self, mass: float, half: bool = True):
        df = phys_calc.dopplerfactor(mass, self.df['InitEnergy'] * 1000)
        coeff = df + 1 / df
        if half == True: coeff /= 2
        self.df['Wavenumber'] = coeff * self.df['Wavenumber']
        self.df.drop(columns=['InitEnergy'], inplace=True)
    
    def wavenumber_cut(self, start: float, end: float):
        """
        This is a method to cut the DataFrame by wavenumber.

        Parameters:
        - start: Start of the wavenumber range (cm^-1)
        - end: End of the wavenumber range (cm^-1)
        """
        self.df = self.df[(self.df['Wavenumber'] >= start) & (self.df['Wavenumber'] <= end)]

    def wavenumber_cut_by_range(self, ranges: list):
        """
        This is a method to cut the DataFrame by multiple wavenumber ranges.

        Parameters:
        - ranges: List of wavenumber ranges, e.g., [[2, 5], [6, 7]]
        """
        mask = pd.Series([False] * len(self.df))
        for start, end in ranges:
            mask |= (self.df['Wavenumber'] >= start) & (self.df['Wavenumber'] <= end)
        self.df = self.df[mask]

    def tof_cut(self, start: float, end: float):
        """
        This is a method to cut the DataFrame by TOF.

        Parameters:
        - start: Start of the TOF range (μs)
        - end: End of the TOF range (μs)
        """
        self.df = self.df[((self.df['TOF'] >= start * 2000) & (self.df['TOF'] <= end * 2000)) | (self.df['TOF'] == -1)]
    
    def channel_cut(self, channel: list = [1,2]):
        """
        This is a method to cut the DataFrame by channel.

        Parameters:
        - channel: List of channels to keep
        """
        self.df = self.df[self.df['Channel'].isin(channel+[-1])]
    
    def timestamp_cut(self, start: float, end: float):
        """
        This is a method to cut the DataFrame by timestamp.

        Parameters:
        - start: Start of the timestamp range (s)
        - end: End of the timestamp range (s)
        """
        self.df = self.df[(self.df['Timestamp'] >= start) & (self.df['Timestamp'] <= end)]

    def draw_tof(self, bins: int = 100):
        """
        This is a method to draw the TOF distribution.

        Parameters:
        - bins: Number of bins for the histogram
        """
        plt.hist(self.df[self.df['TOF'] != -1]['TOF'] / 2000, bins=bins)
        plt.show()

    def draw_hist2d(self, tof_bins: int = 100, hfs_bins: int = 100):
        plt.hist2d(x = self.df[self.df['TOF'] != -1]['Wavenumber'], y = self.df[self.df['TOF'] != -1]['TOF'] / 2000, bins = [hfs_bins, tof_bins])
        plt.show()

    def count_rate(self, bin_width: float = 20, bunch_per_second: float = 100 ,is_draw: bool = True, save_path: str | None = None) -> pd.DataFrame:
        """
        This is a method to calculate the count rate and error, and return the DataFrame.

        Parameters:
        - bin_width: Width of the bin (MHz)
        - bunch_per_second: Bunches per second
        """

        start = self.df['Wavenumber'].min()
        end = self.df['Wavenumber'].max()
        bin_num = int((end - start) / (bin_width * phys_calc.MHz_to_invcm)) + 1
        end = start + bin_num * bin_width * phys_calc.MHz_to_invcm

        bins = np.linspace(start, end, bin_num + 1)

        bunches = self.df[self.df['Channel'] == -1].copy()
        counts = self.df[self.df['Channel'] != -1].copy()
        bunches['binned'] = pd.cut(bunches['Wavenumber'], bins=bins, include_lowest=True)
        counts['binned'] = pd.cut(counts['Wavenumber'], bins=bins, include_lowest=True)
        
        # bunches.to_csv('bunches.csv')
        # counts.to_csv('counts.csv')

        bin_bunches = bunches['binned'].value_counts().sort_index()
        bin_counts = counts['binned'].value_counts().sort_index()

        rates = pd.merge(bin_bunches, bin_counts, left_index=True, right_index=True, how='left')
        rates.columns = ['Bunch', 'Count']
        rates.index = rates.index.map(lambda x: x.mid * phys_calc.invcm_to_MHz)

        rates = rates[rates['Bunch'] > 0]
        rates['y'] = rates['Count'] / rates['Bunch'] * bunch_per_second
        rates['yerr'] = rates['Count'] ** 0.5 / rates['Bunch'] * bunch_per_second
        rates.loc[rates['yerr'] == 0, 'yerr'] += 1 / rates['Bunch'] * bunch_per_second # 防止除零错误

        if save_path is not None:
            rates.to_csv(save_path)

        if is_draw == True:
            plt.errorbar(rates.index, rates['y'], yerr=rates['yerr'], fmt='o', color='black', markersize=3)
            plt.ylabel('Count Rate (cps)')
            plt.xlabel('Wavenumber (invcm)')
            plt.title('Count Rate vs. Wavenumber')
            plt.show()

        return rates

class HFS_fit:
    def __init__(self, data: pd.DataFrame, fit_ini: dict | None = None):
        '''
        Parameters:
        - data: HFS_data object
        - fit_ini: Initial guess for the fit parameters: I, J, ABC, transition wavenumber in invcm
        - df: Shift of the data in MHz
        - fwhmg: FWHM of the Gaussian in MHz
        - fwhml: FWHM of the Lorentzian in MHz
        - scale: Scaling factor for the fit
        - bg: Background level for the fit

        Example:
            fit_ini = {
                "I": 0,
                "J": [0, 0],
                "ABC": [0, 0, 0, 0, 0, 0],
                "trans": 0
            }
        '''
        self.data = data
        self.x = np.array(self.data.index)
        self.y = np.array(self.data['y'])
        self.yerr = np.array(self.data['yerr'])

        if fit_ini:
            self.fit_ini = fit_ini
        else:
            self.fit_ini = {"I": 0, "J": [0, 0], "ABC": [0, 0, 0, 0, 0, 0], "trans": 0}

    def import_json(self, file_path: str):
        import json
        with open(file_path, 'r') as file:
            self.fit_ini = json.load(file)
            if 'I' not in self.fit_ini:
                self.fit_ini['I'] = 0
            if 'J' not in self.fit_ini:
                self.fit_ini['J'] = [0, 0]
            if 'ABC' not in self.fit_ini:
                self.fit_ini['ABC'] = [0, 0, 0, 0, 0, 0]
            if 'trans' not in self.fit_ini:
                self.fit_ini['trans'] = 0

    def fit_with_satlas1(self, shape: str, df: float = 0, fwhm: float = 30, scale: float = 1, bg: list | float = [0], is_fit: bool = True, is_AB_fixed: bool = False, Au_Al_ratio: float | None = None, params: dict = { 'a': -0.25}, boundaries: dict | None = None):
        import satlas as sat
        x = self.x - self.fit_ini['trans'] * phys_calc.invcm_to_MHz

        if isinstance(bg, float) or isinstance(bg, int): bg = [bg]

        if shape == 'crystalball':
            s_main = sat.HFSModel(self.fit_ini['I'], self.fit_ini['J'], self.fit_ini['ABC'], df, fwhm, scale, bg, shape='crystalball', crystalballparams=params)
        elif shape == 'asymmlorentzian':
            s_main = sat.HFSModel(self.fit_ini['I'], self.fit_ini['J'], self.fit_ini['ABC'], df, fwhm, scale, bg, shape='asymmlorentzian', asymmetryparams=params)
        
        if is_AB_fixed:        
            s_main.params['Au'].vary = False
            s_main.params['Al'].vary = False
            s_main.params['Bu'].vary = False
            s_main.params['Bl'].vary = False
        s_main.params['Cu'].vary = False
        s_main.params['Cl'].vary = False

        if boundaries is not None: s_main.set_boundaries(boundaries)

        if Au_Al_ratio is not None: s_main.fix_ratio(Au_Al_ratio, target='upper', parameter='A')
        
        if is_fit:
            sat.chisquare_fit(s_main, x, self.y, self.yerr)
            s_main.display_chisquare_fit()
            self.fit_para_result = s_main.get_result_dict()

        self.fit_result_x = np.linspace(min(x), max(x), 5000)
        self.fit_result_y = s_main(self.fit_result_x)

    def crystalball_fit(self, df: float = 0, fwhm: float = 30, scale: float = 1, bg: float = 0, is_fit: bool = True, is_AB_fixed: bool = False, Au_Al_ratio: float | None = None, crystalballparams: dict = { 'Taillocation': -0.25,'Tailamplitude': 6}, boundaries: dict | None = None):
        # import satlas as sat
        # x = self.x - self.fit_ini['trans'] * phys_calc.invcm_to_MHz

        # s_main = sat.HFSModel(self.fit_ini['I'], self.fit_ini['J'], self.fit_ini['ABC'], df, fwhm, scale, [bg], shape='crystalball', crystalballparams=crystalballparams)
        # if is_AB_fixed:        
        #     s_main.params['Au'].vary = False
        #     s_main.params['Al'].vary = False
        #     s_main.params['Bu'].vary = False
        #     s_main.params['Bl'].vary = False
        # s_main.params['Cu'].vary = False
        # s_main.params['Cl'].vary = False

        # if boundaries is not None: s_main.set_boundaries(boundaries)

        # if Au_Al_ratio is not None: s_main.fix_ratio(Au_Al_ratio, target='upper', parameter='A')
        
        # if is_fit:
        #     sat.chisquare_fit(s_main, x, self.y, self.yerr)
        #     s_main.display_chisquare_fit()
        #     self.fit_para_result = s_main.get_result_dict()

        # self.fit_result_x = np.linspace(min(x), max(x), 1000)
        # self.fit_result_y = s_main(self.fit_result_x)
        self.fit_with_satlas1('crystalball', df, fwhm, scale, bg, is_fit, is_AB_fixed, Au_Al_ratio, crystalballparams, boundaries)
    
    def asymmlorentzian_fit(self, df: float = 0, fwhm: float = 30, scale: float = 1, bg: float = 0, is_fit: bool = True, is_AB_fixed: bool = False, Au_Al_ratio: float | None = None, asymmetryparams: dict = { 'a': -0.25}, boundaries: dict | None = None):
        # import satlas as sat
        # x = self.x - self.fit_ini['trans'] * phys_calc.invcm_to_MHz

        # s_main = sat.HFSModel(self.fit_ini['I'], self.fit_ini['J'], self.fit_ini['ABC'], df, fwhm, scale, [bg], shape='asymmlorentzian', asymmetryparams=asymmetryparams)
        # if is_AB_fixed:        
        #     s_main.params['Au'].vary = False
        #     s_main.params['Al'].vary = False
        #     s_main.params['Bu'].vary = False
        #     s_main.params['Bl'].vary = False
        # s_main.params['Cu'].vary = False
        # s_main.params['Cl'].vary = False

        # if boundaries is not None: s_main.set_boundaries(boundaries)

        # if Au_Al_ratio is not None: s_main.fix_ratio(Au_Al_ratio, target='upper', parameter='A')
        
        # if is_fit:
        #     sat.chisquare_fit(s_main, x, self.y, self.yerr)
        #     s_main.display_chisquare_fit()
        #     self.fit_para_result = s_main.get_result_dict()

        # self.fit_result_x = np.linspace(min(x), max(x), 1000)
        # self.fit_result_y = s_main(self.fit_result_x)
        self.fit_with_satlas1('asymmlorentzian', df, fwhm, scale, bg, is_fit, is_AB_fixed, Au_Al_ratio, asymmetryparams, boundaries)
    
    def voigt_fit(self, df: float = 0, fwhmg: float = 30, fwhml: float = 20, scale: float = 1, bg: float = 0, is_fit: bool = True, is_AB_fixed: bool = False, is_B_fixed: bool = False, use_racah: bool = False, Au_Al_ratio: float | None = None, param_prior: dict | None = None, sidepeak_params: dict | None = None):
        """
        This is a method to fit the data with Voigt profile using satlas2.
        """
        import satlas2 as sat
        x = self.x - self.fit_ini['trans'] * phys_calc.invcm_to_MHz

        datasource = sat.Source(x, self.y, yerr=self.yerr, name='Data')
        f = sat.Fitter()

        if sidepeak_params is not None:
            s_main = sat.HFS(self.fit_ini['I'], self.fit_ini['J'], self.fit_ini['ABC'][:2], self.fit_ini['ABC'][2:4], self.fit_ini['ABC'][4:],df = df, fwhmg=fwhmg, fwhml=fwhml, name='main', scale=scale, racah=use_racah, N = sidepeak_params['N'], offset = sidepeak_params['offset'], poisson = sidepeak_params['poisson'])
        else:
            s_main = sat.HFS(self.fit_ini['I'], self.fit_ini['J'], self.fit_ini['ABC'][:2], self.fit_ini['ABC'][2:4], self.fit_ini['ABC'][4:],df = df, fwhmg=fwhmg, fwhml=fwhml, name='main', scale=scale, racah=use_racah)

        background = sat.Polynomial([bg], 'bg')

        if is_AB_fixed:        
            s_main.params['Au'].vary = False
            s_main.params['Al'].vary = False
            s_main.params['Bu'].vary = False
            s_main.params['Bl'].vary = False
        s_main.params['Cu'].vary = False
        s_main.params['Cl'].vary = False
        # s_main.params['FWHMG'].vary = False

        if is_B_fixed:
            s_main.params['Bu'].vary = False
            s_main.params['Bl'].vary = False
        
        datasource.addModel(s_main)
        datasource.addModel(background)
        f.addSource(datasource)

        if Au_Al_ratio is not None: f.setExpr(["Data___main___Au"], "Data___main___Al * " + str(Au_Al_ratio))
        
        if param_prior is not None: 
            for key in param_prior:
                f.setParamPrior('Data', 'main', key, param_prior[key][0], param_prior[key][1])

        if is_fit:
            f.fit()
            print(f.reportFit())
            self.fit_para_result = f.createResultDataframe()

        self.fit_result_x = np.linspace(min(x), max(x), 5000)
        self.fit_result_y = datasource.evaluate(self.fit_result_x)

    def draw(self):
        fig, ax = plt.subplots(figsize=(8, 6))
        ax.plot(self.fit_result_x, self.fit_result_y, '-',lw=1., c= 'r', label='Fitting')
        ax.errorbar(self.x - self.fit_ini['trans'] * phys_calc.invcm_to_MHz, self.y, self.yerr, fmt='ko', markersize=3., label='Data')
        ax.set_xlabel("Relative Frequency (MHz)")
        ax.set_ylabel("Rate (cps)")
        ax.legend()
        plt.show()

    def brokenaxes_draw(self, range: tuple, y_log: bool = False, errorbar_alpha: float = 1):

        fig = plt.figure(figsize=(8, 6))
        bax = brokenaxes.brokenaxes(xlims=range)

        bax.plot(self.fit_result_x, self.fit_result_y, '-',lw=1., c= 'r', label='Fitting')
        # 绘图
        bax.errorbar(self.x - self.fit_ini['trans'] * phys_calc.invcm_to_MHz, self.y, self.yerr, fmt='ko', markersize=3., label='Data', alpha=errorbar_alpha)
        if y_log:
            bax.set_yscale('log')
        bax.set_xlabel("Relative Frequency (MHz)")
        bax.set_ylabel("Rate (cps)")

        bax.legend()
        plt.show()

class HFS_simulation:
    def __init__(self, source_yield: float, efficiency: float, background_ratio: float, time: float, bin_num: int, scan_range: list, fit_ini: dict | None = None):
        import satlas2 as sat
        if fit_ini:
            self.para = fit_ini
        else:
            self.para = {"I": 0, "J": [0, 0], "ABC": [0, 0, 0, 0, 0, 0], "fwhm": 30}

        rate = source_yield * efficiency
        self.model = sat.HFSModel(self.para['I'], self.para['J'], self.para['ABC'], 0, self.para['fwhm'],
                    scale=rate * time / bin_num,
                    background_params=[background_ratio * rate * time / bin_num],
                    use_racah=True)
        
        x = np.linspace(scan_range[0], scan_range[1], 1000)
        y = self.model.f(x)

        data_x = np.linspace(scan_range[0], scan_range[1], bin_num)

        data_y = sat.generateSpectrum(self.model, data_x)
        yerr = np.sqrt(data_y)

        fig = plt.figure()
        ax = fig.add_axes([0.1, 0.1, 0.8, 0.8])
        ax.errorbar(data_x, data_y, yerr=np.sqrt(data_y), fmt='o', label='Data')

        ax.plot(x, y, label='Initial guess')
        ax.set_xlabel('Frequency (MHz)')
        ax.set_ylabel('Counts')

        datasource = sat.Source(data_x, data_y, yerr=yerr, name='Datafile1')
        datasource.addModel(self.model)
        # f = sat.Fitter()
        # f.addSource(datasource)
        # f.fit()
        # print(f.reportFit())

        # ax.plot(x, self.model.f(x), label='Fit')
        # ax.legend(loc=0)
        ax.legend(loc=0)

        self.x = data_x
        self.y = data_y
        self.yerr = yerr

        plt.show()

    def get_result(self):
        return self.x, self.y, self.yerr


"""
Implementation of classes for different lineshapes, creating callables for easy and intuitive calculations.

.. moduleauthor:: Wouter Gins <wouter.gins@kuleuven.be>
.. moduleauthor:: Ruben de Groote <ruben.degroote@kuleuven.be>
.. moduleauthor:: Kara Marie Lynch <kara.marie.lynch@cern.ch>
"""
import numpy as np
from scipy.special import wofz

__all__ = ['Gaussian', 'Lorentzian', 'Voigt', 'PseudoVoigt', 'Crystalball', 'AsymmLorentzian']
sqrt2 = 2 ** 0.5
sqrt2pi = (2 * np.pi) ** 0.5
sqrt2log2t2 = 2 * np.sqrt(2 * np.log(2))
base_e = np.exp(1)


class Profile(object):
    """Abstract baseclass for defining lineshapes."""

    def __init__(self, fwhm=None, mu=None, amp=None, ampIsArea=False):
        super(Profile, self).__init__()
        self.ampIsArea = ampIsArea
        self.fwhm = np.abs(fwhm) if fwhm is not None else np.abs(1.0)
        self.mu = mu if mu is not None else 0.0
        self.amp = amp if amp is not None else 1.0

    @property
    def mu(self):
        """Location of the peak."""
        return self._mu

    @mu.setter
    def mu(self, value):
        self._mu = value

    @property
    def fwhm(self):
        """FWHM of the peak."""
        return self._fwhm

    @fwhm.setter
    def fwhm(self, value):
        self._fwhm = value

    @property
    def amp(self):
        """Peak amplitude."""
        return self._amp

    @amp.setter
    def amp(self, value):
        self._amp = value

    @property
    def ampIsArea(self):
        """Boolean controlling the behaviour of *amp*."""
        return self._ampIsArea

    @ampIsArea.setter
    def ampIsArea(self, value):
        self._ampIsArea = value

    def __repr__(self):
        s = str(type(self)) + 'FWHM: {}, mu: {}, amp: {}'
        s = s.format(self.fwhm, self.mu, self.amp)
        return s

    def __call__(self, vals):
        """Evaluates the lineshape in the given values.

        Parameters
        ----------
        vals: array_like
            Array of values to evaluate the lineshape in.

        Returns
        -------
        array_like
            Array of seperate response values of the lineshape."""
        if self.ampIsArea:
            factor = 1.0
        else:
            factor = self._normFactor
        vals = vals / factor
        return self.amp * vals

class Gaussian(Profile):

    r"""A callable normalized Gaussian profile."""

    def __init__(self, fwhm=None, mu=None, amp=None, ampIsArea=False):
        """Creates a callable object storing the fwhm, amplitude and location
        of a Gaussian lineshape.

        Parameters
        ----------
        fwhm: float
            Full Width At Half Maximum, defaults to 1.
        mu: float
            Location of the center, defaults to 0.
        amp: float
            Amplitude of the profile, defaults to 1.
        ampIsArea: boolean
            Sets if the amplitude is the integral or the peakheight. Defaults
            to False.

        Returns
        -------
        Gaussian
            Callable instance, evaluates the Gaussian profile in the arguments
            supplied."""
        super(Gaussian, self).__init__(fwhm=fwhm, mu=mu,
                                       amp=amp, ampIsArea=ampIsArea)

    @property
    def fwhm(self):
        """FWHM of the peak."""
        return self._fwhm

    @fwhm.setter
    def fwhm(self, value):
        self._fwhm = value
        self.sigma = self.fwhm / (sqrt2log2t2)
        if not self.ampIsArea:
            self._normFactor = (self.sigma * sqrt2pi) ** (-1)

    def __call__(self, x):
        r"""Evaluates the lineshape in the given values.

        Parameters
        ----------
        vals: array_like
            Array of values to evaluate the lineshape in.

        Returns
        -------
        array_like
            Array of seperate response values of the lineshape.

        Note
        ----
        The used formula is taken from the MathWorld webpage
        http://mathworld.wolfram.com/GaussianFunction.html:

            .. math::
                G(x;\mu, \sigma) &= \frac{\exp\left(-\frac{1}{2}\left(\frac{x-\mu}
                {\sigma}\right)^2\right)}{\sqrt{2\pi}\sigma}

                FWHM &= s\sqrt{2\ln\left(2\right)}\sigma"""
        x = x - self.mu
        expPart = np.exp(-0.5 * (x / self.sigma) ** 2)
        normPart = self.sigma * sqrt2pi
        return super(Gaussian, self).__call__(expPart / normPart)


class Lorentzian(Profile):

    """A callable normalized Lorentzian profile."""

    def __init__(self, fwhm=None, mu=None, amp=None, ampIsArea=False):
        """Creates a callable object storing the fwhm, amplitude and location
        of a Lorentzian lineshape.

        Parameters
        ----------
        fwhm: float
            Full Width At Half Maximum, defaults to 1.
        mu: float
            Location of the center, defaults to 0.
        amp: float
            Amplitude of the profile, defaults to 1.
        ampIsArea: boolean
            Sets if the amplitude is the integral or the peakheight. Defaults
            to False.

        Returns
        -------
        Lorentzian
            Callable instance, evaluates the Lorentzian profile in the arguments
            supplied."""
        super(Lorentzian, self).__init__(fwhm=fwhm, mu=mu,
                                         amp=amp, ampIsArea=ampIsArea)

    @property
    def fwhm(self):
        """FWHM of the peak."""
        return self._fwhm

    @fwhm.setter
    def fwhm(self, value):
        self._fwhm = value
        self.gamma = 0.5 * self.fwhm
        if not self.ampIsArea:
            self._normFactor = 1.0 / (self.gamma * np.pi)

    def __call__(self, x):
        r"""Evaluates the lineshape in the given values.

        Parameters
        ----------
        vals: array_like
            Array of values to evaluate the lineshape in.

        Returns
        -------
        array_like
            Array of seperate response values of the lineshape.

        Note
        ----
        The formula used is taken from the MathWorld webpage
        http://mathworld.wolfram.com/LorentzianFunction.html:

            .. math::
                \mathcal{L}\left(x; \mu, \gamma\right) &= \frac{\gamma}
                {\pi\left(\left(x-\mu\right)^2+\gamma^2\right)}

                FWHM &= 2\gamma"""
        x = x - self.mu
        topPart = self.gamma
        bottomPart = (x ** 2 + self.gamma ** 2) * np.pi
        return super(Lorentzian, self).__call__(topPart / bottomPart)

class AsymmLorentzian(Profile):

    """A callable normalized Lorentzian profile with builtin asymmetry."""

    def __init__(self, fwhm=None, mu=None, amp=None, ampIsArea=False, asymm=0):
        """Creates a callable object storing the fwhm, amplitude, location and asymmetry
        of a asymmetric Lorentzian lineshape.

        Parameters
        ----------
        fwhm: float
            Full Width At Half Maximum, defaults to 1.
        mu: float
            Location of the center, defaults to 0.
        amp: float
            Amplitude of the profile, defaults to 1.
        ampIsArea: boolean
            Sets if the amplitude is the integral or the peakheight. Defaults
            to False.
        asymm: float
            Asymmetry parameter, default to 0.

        Returns
        -------
        AsymmLorentzian
            Callable instance, evaluates the profile in the arguments supplied."""
        self.asymm = asymm
        super(AsymmLorentzian, self).__init__(fwhm=fwhm, mu=mu,
                                              amp=amp, ampIsArea=ampIsArea)

    @property
    def fwhm(self):
        """FWHM of the peak."""
        return self._fwhm

    @fwhm.setter
    def fwhm(self, value):
        self._fwhm = value
        self.gamma = 0.5 * self.fwhm
        if not self.ampIsArea:
            self._normFactor = 1.0 / (self.gamma * np.pi)

    @property
    def asymm(self):
        """Asymmetry of the profile."""
        return self._asymm

    @asymm.setter
    def asymm(self, value):
        self._asymm = value

    def linewidth_function(self, x):
        r"""Calculates the FWHM."""
        return 2 * self.gamma / (1 + np.exp(self.asymm*x))

    def __call__(self, x):
        r"""Evaluates the lineshape in the given values.

        Parameters
        ----------
        vals: array_like
            Array of values to evaluate the lineshape in.

        Returns
        -------
        array_like
            Array of seperate response values of the lineshape.

        Note
        ----
        The formula used is taken from the MathWorld webpage
        http://mathworld.wolfram.com/LorentzianFunction.html and an empirical
        linewidth function:

            .. math::
                \mathcal{L}\left(x; \mu, \gamma\right) &= \frac{\gamma}
                {\pi\left(\left(x-\mu\right)^2+\gamma^2\right)}

                FWHM &= \frac{2\gamma}{1+\exp\left(asymm \left(x-\mu\right)\right)}"""
        x = x - self.mu
        gamma = self.linewidth_function(x)
        topPart = gamma
        bottomPart = (x ** 2 + gamma ** 2) * np.pi
        return super(AsymmLorentzian, self).__call__(topPart / bottomPart)

class PseudoVoigt(Profile):

    r"""A callable normalized PseudoVoigt profile.

    Parameters
    ----------
    fwhm: float
        Full Width At Half Maximum, defaults to 1.
    mu: float
        Location of the center, defaults to 0.
    amp: float
        Amplitude of the profile, defaults to 1.

    Returns
    -------
    PseudoVoigt
        Callable instance, evaluates the pseudovoigt profile in the arguments
        supplied.

    Note
    ----
    The formula used is taken from the webpage
    http://en.wikipedia.org/wiki/Voigt_profile#Pseudo-Voigt_Approximation, and the
    supplied FWHM is appropriately transformed for the Gaussian and Lorentzian
    lineshapes:

    .. math::
        \mathcal{V}\left(x; \mu, \eta, \sigma, \gamma\right) = \eta \mathcal{L}
        (x; \sigma, \mu) + (1-\eta) G(x; \sigma, \mu)

    The formula is also adapted to incorporate an asymmetry in the lineshape. This
    is done by varying the used fwhm with a sigmoid function over the frequency range."""

    def __init__(self, eta=None, fwhm=None, mu=None,
                 amp=None, **kwargs):
        self.L = Lorentzian(**kwargs)
        self.G = Gaussian(**kwargs)
        self._a = 0.
        self._n = np.abs(eta) if eta is not None else 0.5
        if self._n > 1:
            self._n = self._n - int(self._n)
        super(PseudoVoigt, self).__init__(fwhm=fwhm, mu=mu,
                                          amp=amp, **kwargs)

    @property
    def fwhm(self):
        """FWHM of the peak."""
        return self._fwhm

    @fwhm.setter
    def fwhm(self, value):
        self._fwhm = value
        self.L.fwhm = value
        self.G.fwhm = value
        if not self.ampIsArea:
            self._normFactor = self.n * self.L(0)
            self._normFactor += (1.0 - self.n) * self.G(0)

    @property
    def n(self):
        """Ratio of Lorentzian aspect."""
        return self._n

    @n.setter
    def n(self, value):
        value = np.abs(value)
        if value > 1:
            value = value - int(value)
        self._n = value
        if not self.ampIsArea:
            self._normFactor = self.n * self.L(0)
            self._normFactor += (1.0 - self.n) * self.G(0)

    @property
    def a(self):
        """Asymmetry parameter."""
        return self._a

    @a.setter
    def a(self, value):
        self._a = value

    def __call__(self, x):
        x = x - self.mu
        fwhm_scale = 2 / (1 + np.exp(self.a * x / self.fwhm))
        self.L.fwhm = self.fwhm*fwhm_scale
        self.G.fwhm = self.fwhm*fwhm_scale

        val = self.n * self.L(x) + (1.0 - self.n) * self.G(x)

        return super(PseudoVoigt, self).__call__(val)

class Voigt(Profile):

    r"""A callable normalized Voigt profile."""

    def __init__(self, fwhm=None, mu=None, amp=None, ampIsArea=False):
        """Creates a callable object storing the fwhm, amplitude and location
        of a Voigt lineshape.

        Parameters
        ----------
        fwhm: float
            Full Width At Half Maximum, defaults to 1.
        mu: float
            Location of the center, defaults to 0.
        amp: float
            Amplitude of the profile, defaults to 1.
        ampIsArea: boolean
            Sets if the amplitude is the integral or the peakheight. Defaults
            to False.

        Returns
        -------
        Voigt
            Callable instance, evaluates the Voigt profile in the arguments supplied."""
        self._fwhmNorm = np.array([sqrt2log2t2, 2])
        super(Voigt, self).__init__(fwhm=fwhm, mu=mu,
                                    amp=amp, ampIsArea=ampIsArea)

    @property
    def fwhm(self):
        """FWHM of the peak."""
        return self._fwhm

    @fwhm.setter
    def fwhm(self, value):
        if isinstance(value, (list, tuple, np.ndarray)):
            seperate = value[0:2]
            self.fwhmG, self.fwhmL = seperate
            G, L = seperate
            self._fwhm = 0.5346 * self.fwhmL + \
                         (0.2166 * self.fwhmL ** 2 + self.fwhmG ** 2) ** 0.5
            self.sigma, self.gamma = seperate / self._fwhmNorm
        else:
            self.fwhmG, self.fwhmL = value, value
            self._fwhm = 0.6144031129489123 * value
            self.sigma, self.gamma = self._fwhm / self._fwhmNorm

        if not self.ampIsArea:
            z = (0 + 1j * self.gamma) / (self.sigma * sqrt2)
            top = wofz(z).real / (self.sigma * sqrt2pi)
            self._normFactor = top

    def __call__(self, x):
        r"""Evaluates the lineshape in the given values.

        Parameters
        ----------
        vals: array_like
            Array of values to evaluate the lineshape in.

        Returns
        -------
        array_like
            Array of seperate response values of the lineshape.

        Note
        ----
        The formula used is taken from the Wikipedia webpage
        http://en.wikipedia.org/wiki/Voigt_profile, with :math:`w(z)` the Faddeeva
        function, and the values supplied as FWHM are appropriately transformed to
        :math:`\sigma` and :math:`\gamma`:

            .. math::
                V\left(x;\mu, \sigma, \gamma\right) &= \frac{\Re\left[w\left(z\right)
                \right]}{\sigma\sqrt{2\pi}}

                z&=\frac{x+i\gamma}{\sigma\sqrt{2\pi}}"""
        x = x - self.mu
        z = (x + 1j * self.gamma) / (self.sigma * sqrt2)
        top = wofz(z).real / (self.sigma * sqrt2pi)
        return super(Voigt, self).__call__(top)

class Crystalball(Profile):

    r"""A callable Crystalball profile."""

    def __init__(self, fwhm=None, mu=None, amp=None, alpha=None, n=None, ampIsArea=False):
        """Creates a callable object storing the fwhm, amplitude and location
        of a Crystalball lineshape. For more information, see :cite:`Gaiser`
        and :cite:`skwarnicki`

        Parameters
        ----------
        fwhm: float
            Full Width At Half Maximum, defaults to 1.
        mu: float
            Location of the center, defaults to 0.
        amp: float
            Amplitude of the profile, defaults to 1.
        alpha: float
            Location of the tail.
        n: float
            Relative amplitude of the tail.
        ampIsArea: boolean
            Sets if the amplitude is the integral or the peakheight. Defaults
            to False.

        Returns
        -------
        Crystalball
            Callable instance, evaluates the Crystalball profile in the arguments supplied."""
        super(Crystalball, self).__init__(fwhm=fwhm, mu=mu,
                                          amp=amp, ampIsArea=ampIsArea)
        self.alpha = alpha
        self.n = n

    @property
    def mu(self):
        return self._mu

    @mu.setter
    def mu(self, value):
        self._mu = value

    @property
    def fwhm(self):
        """FWHM of the peak."""
        return self._fwhm

    @fwhm.setter
    def fwhm(self, value):
        self._fwhm = value
        self.sigma = self.fwhm / (sqrt2log2t2)

        if not self.ampIsArea:
            self._normFactor = 1

    @property
    def alpha(self):
        return self._alpha

    @alpha.setter
    def alpha(self, value):
        self._alpha = value

    @property
    def n(self):
        return self._n

    @n.setter
    def n(self, value):
        self._n = value

    def _bigger(self, x):
        ret = np.exp(-0.5 * x * x)
        return ret

    def _smaller(self, x):
        a = np.abs(self.alpha)
        n = self.n
        b = n / a - a
        a = ((n / a)**n) * np.exp(-0.5*a*a)
        return a / (b - x) ** n

    def __call__(self, x):
        r"""Evaluates the lineshape in the given values.

        Parameters
        ----------
        vals: array_like
            Array of values to evaluate the lineshape in.

        Returns
        -------
        array_like
            Array of seperate response values of the lineshape."""
        x = (x - self.mu) * np.sign(self.alpha) / self.sigma
        y = np.piecewise(x, x >= -np.abs(self.alpha), [self._bigger, self._smaller])
        return super(Crystalball, self).__call__(y)



"""
Implementation of calculation of the loglikelihood for common distributions.

.. moduleauthor:: Wouter Gins <wouter.gins@kuleuven.be>
"""
import numpy as np
import scipy as sp

sqrt2pi = np.sqrt(2*np.pi)
__all__ = ['poisson_llh', 'create_gaussian_llh' , 'create_gaussian_priormap']


def poisson_llh(y, f, x):
    """Returns the loglikelihood for a Poisson distribution.
    In this calculation, it is assumed that the parameters
    are true, and the loglikelihood that the data is drawn from
    the distribution established by the parameters is calculated.

    Parameters
    ----------
    y : array_like
        Data to which is being fitted.
    l : array_like
        Result from the model.

    Returns
    -------
    array_like
        Array with loglikelihoods for the data."""
    l = np.hstack(f(x))
    return y * np.log(l) - l

def create_gaussian_llh(yerr=1, xerr=None, func=None):
    """Returns the loglikelihood-function for a Gaussian distribution,
    with the given uncertainty on the data points. The input parameters
    will be (in order) the data to be fitted and the model response.

    Parameters
    ----------
    yerr : array_like
        Measured uncertainties on the datapoint.

    Returns
    -------
    function
        Function that calculates the loglikelihood for the given data and model values."""

    if func is not None:
        if xerr is not None:
            def gaussian_llh(y, f, x, xerr=xerr):
                l = np.hstack(np.array([f(x)]))
                yerr = func(l)
                xerr = np.hstack(np.array([(sp.misc.derivative(f, x, dx=1E-6) * xerr)]))
                bottom = np.sqrt(yerr * yerr + xerr * xerr)
                return -0.5*( (y - l) / bottom)**2
            return gaussian_llh
        else:
            def gaussian_llh(y, f, x):
                l = np.hstack(np.array([f(x)]))
                bottom = func(l)
                return -0.5*( (y - l) / bottom)**2
            return gaussian_llh
    else:
        if xerr is not None:
            def gaussian_llh(y, f, x, xerr=xerr, yerr=yerr):
                l = f(x)
                xerr = np.hstack(np.array([(sp.misc.derivative(f, x, dx=1E-6) * xerr)]))
                bottom = np.sqrt(yerr * yerr + xerr * xerr)
                return -0.5*( (y - l) / bottom)**2
            return gaussian_llh
        else:
            def gaussian_llh(y, f, x, yerr=yerr):
                l = np.hstack(np.array([f(x)]))
                return -0.5*( (y - l) / yerr)**2
            return gaussian_llh

def create_gaussian_priormap(literature_value, uncertainty):
    """Generates a function that describes a Gaussian prior mapping around
    the given literature value with the given uncertainty.

    Parameters
    ----------
    literature_value : float
        Value for the parameter which is optimal.
    uncertainty : float
        Value for the uncertainty on the parameter.

    Returns
    -------
    function
        Function that calculates the prior value for the given
        parameter value."""
    def func(value):
        """Returns the Gaussian prior with center {:.2f}
        and {:.2f} sigma.

        Parameters
        ----------
        value : float
            Current value of the parameter.

        Returns
        -------
        float
            Value of the prior.""".format(literature_value, uncertainty)
        deviation = (value - literature_value) / uncertainty
        return -0.5 * deviation * deviation
    return func



"""
Implementation of fitting routines specialised for BaseModel objects. Note that not all functions are loaded into the global satlas namespace.

.. moduleauthor:: Wouter Gins <wouter.gins@kuleuven.be>
.. moduleauthor:: Ruben de Groote <ruben.degroote@kuleuven.be>
"""
import copy
import os
import numdifftools as nd

from satlas.stats import emcee as mcmc

import lmfit as lm
from satlas import loglikelihood as llh
from satlas import tqdm
import h5py
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from scipy import optimize
# from scipy.misc import derivative
from scipy.stats import chi2

from findiff import Diff

__all__ = ['chisquare_spectroscopic_fit', 'chisquare_fit', 'calculate_analytical_uncertainty',
           'likelihood_fit', 'likelihood_walk', 'create_band', 'process_walk']
chisquare_warning_message = "The supplied dictionary for {} did not contain the necessary keys 'value' and 'uncertainty'."

###############################
# CHI SQUARE FITTING ROUTINES #
###############################

def chisquare_model(params, f, x, y, yerr, xerr=None, func=None):
    r"""Model function for chisquare fitting routines as established
    in this module.

    Parameters
    ----------
    params: lmfit.Parameters
        Instance of lmfit.Parameters object, to be assigned to the model object.
    f: :class:`.BaseModel`
        Callable instance with the correct methods for the fitmethods.
    x: array_like
        Experimental data for the x-axis.
    y: array_like
        Experimental data for the y-axis.
    yerr: array_like
        Experimental errorbars on the y-axis.

    Other parameters
    ----------------
    xerr: array_like, optional
        Given an array with the same size as *x*, the error is taken into
        account by using the method of estimated variance. Defaults to *None*.
    func: function, optional
        Given a function, the errorbars on the y-axis is calculated from
        the fitvalue using this function. Defaults to *None*.

    Returns
    -------
    NumPy array
        Array containing the residuals for the given parameters, divided by the
        uncertainty.

    Note
    ----
    If a custom function is to be used for the calculation of the residual,
    this function should be overwritten.

    The method of estimated variance calculates the chisquare in the following way:

        .. math::

            \sqrt{\chi^2} = \frac{y-f(x)}{\sqrt{\sigma_x^2+f'(x)^2\sigma_x^2}}"""
    f.params = params
    model = np.hstack(f(x))
    if func is not None:
        yerr = func(model)
    if xerr is not None:
        x = np.array(x)
        d_dx = Diff(0, 1E-5)
        if len(x.shape) > 1:
            xerr = d_dx(lambda x: np.hstack(f.seperate_response(x)))(x) * xerr
            # xerr = derivative(lambda x: np.hstack(f.seperate_response(x)), x, dx=1E-5) * xerr
        else:
            xerr = d_dx(f)(x) * xerr
            # xerr = derivative(f, x, dx=1E-5) * xerr
        bottom = np.sqrt(yerr * yerr + xerr * xerr)
    else:
        bottom = yerr
    return_value = (y - model) / bottom
    appended_values = f.get_chisquare_mapping()
    if appended_values is not None:
        return_value = np.append(return_value, appended_values)
    return return_value

def chisquare_spectroscopic_fit(f, x, y, xerr=None, func=np.sqrt, verbose=True, hessian=False, method='leastsq'):
    """Use the :func:`chisquare_fit` function, automatically estimating the errors
    on the counts by the square root.

    Parameters
    ----------
    f: :class:`.BaseModel`
        Model object containing all the information about the fit;
        will be fitted to the given data.
    x: array_like
        Experimental data for the x-axis.
    y: array_like
        Experimental data for the y-axis.

    Other parameters
    ----------------
    xerr: array_like, optional
        Error bars on *x*.
    func: function, optional
        This function is applied to the model values to calculate the errorbars.
    verbose: boolean, optional
        When set to *True*, a tqdm-progressbar in the terminal is maintained.
        Defaults to *True*.
    hessian: boolean, optional
        When set to *True*, the SATLAS implementation of the Hessian uncertainty estimate
        is calculated, otherwise the LMFIT version is used. Defaults to *False*.
    method: string, optional
        Sets the method to be used by lmfit for the fitting. See lmfit for all options.

    Return
    ------
    success, message: tuple
        Boolean indicating the success of the convergence, and the message
        from the optimizer."""
    y = np.hstack(y)
    yerr = np.sqrt(y)
    yerr[np.isclose(yerr, 0.0)] = 1.0
    return chisquare_fit(f, x, y, yerr=yerr, xerr=xerr, func=func, verbose=verbose, hessian=hessian, method=method)

def chisquare_fit(f, x, y, yerr=None, xerr=None, func=None, verbose=True, hessian=False, method='leastsq'):
    """Use a non-linear least squares minimization (Levenberg-Marquardt)
    algorithm to minimize the chi-square of the fit to data *x* and
    *y* with errorbars *yerr*.

    Parameters
    ----------
    f: :class:`.BaseModel`
        Model object containing all the information about the fit;
        will be fitted to the given data.
    x: array_like
        Experimental data for the x-axis.
    y: array_like
        Experimental data for the y-axis.
    yerr: array_like
        Uncertainties on *y*.

    Other parameters
    ----------------
    xerr: array_like, optional
        Uncertainties on *x*.
    func: function, optional
        Uses the provided function on the fitvalue to calculate the
        errorbars.
    verbose: boolean, optional
        When set to *True*, a tqdm-progressbar in the terminal is maintained.
        Defaults to *True*.
    hessian: boolean, optional
        When set to *True*, the SATLAS implementation of the Hessian uncertainty estimate
        is calculated, otherwise the LMFIT version is used. Defaults to *False*.
    method: string, optional
        Sets the method to be used by lmfit for the fitting. See lmfit for all options.

    Return
    ------
    success, message: tuple
        Boolean indicating the success of the convergence, and the message
        from the optimizer."""

    params = f.params

    if verbose:
        def iter_cb(params, iter, resid, *args, **kwargs):
            progress.set_description('Chisquare fitting in progress (' + str((resid**2).sum()) + ')')
            progress.update(1)
        progress = tqdm.tqdm(desc='Chisquare fitting in progress', leave=True)
    else:
         def iter_cb(params, iter, resid, *args, **kwargs):
            pass

    result = lm.minimize(chisquare_model, params, args=(f, x, np.hstack(y), np.hstack(yerr), xerr, func), iter_cb=iter_cb, method=method)
    f.params = copy.deepcopy(result.params)
    f.chisqr_chi = copy.deepcopy(result.chisqr)

    success = False
    counter = 0
    while not success:
        result = lm.minimize(chisquare_model, result.params, args=(f, x, np.hstack(y), np.hstack(yerr), xerr, func), iter_cb=iter_cb, method=method)
        f.params = copy.deepcopy(result.params)
        success = np.isclose(result.chisqr, f.chisqr_chi)
        f.chisqr_chi = copy.deepcopy(result.chisqr)
        if counter > 10 and not success:
            break
    if verbose:
        progress.set_description('Chisquare fitting done')
        progress.close()

    f.ndof_chi = copy.deepcopy(result.nfree)
    f.redchi_chi = copy.deepcopy(result.redchi)
    f.chisq_res_par = copy.deepcopy(f.params)
    f.aic_chi = copy.deepcopy(result.aic)
    f.bic_chi = copy.deepcopy(result.bic)
    if hessian:
        if verbose:
            progress = tqdm.tqdm(desc='Starting Hessian calculation', leave=True, miniters=1)
        else:
            progress = None
        assign_hessian_estimate(lambda *args: (chisquare_model(*args)**2).sum(), f, f.chisq_res_par, x, np.hstack(y), np.hstack(yerr), xerr, func, progress=progress)
    else:
        for key in f.params.keys():
            if f.params[key].stderr is not None:
                f.params[key].stderr /= f.redchi_chi**0.5
                f.chisq_res_par[key].stderr /= f.redchi_chi**0.5

    return success, result.message

##########################################
# MAXIMUM LIKELIHOOD ESTIMATION ROUTINES #
##########################################


class PriorParameter(lm.Parameter):

    # Extended the Parameter class from LMFIT to incorporate prior boundaries.

    def __init__(self, name=None, value=None, vary=True, min=None, max=None,
                 expr=None, priormin=None, priormax=None):
        super(PriorParameter, self).__init__(name=name, value=value,
                                             vary=vary, min=min,
                                             max=max, expr=expr)
        self.priormin = priormin
        self.priormax = priormax

    def __getstate__(self):
        return_value = super(PriorParameter, self).__getstate__()
        return_value += (self.priormin, self.priormax)
        return return_value

    def __setstate__(self, state):
        state_pass = state[:-2]
        self.priormin, self.priormax = state[-2:]
        super(PriorParameter, self).__setstate__(state_pass)

sqrt2pi = np.sqrt(2*np.pi)

def likelihood_x_err(f, x, y, xerr, func):
    """Calculates the loglikelihood for a model given
    x and y values. Incorporates a common given error on
    the x-axis.

    Parameters
    ----------
    f: :class:`.BaseModel`
        Model object set to current parameters.
    x: array_like
        Experimental data for the x-axis.
    y: array_like
        Experimental data for the y-axis.
    xerr: float
        Experimental uncertainty on *x*.
    func: function
        Function taking (*y_data*, *y_model*) as input,
        and returning the loglikelihood that the data is
        drawn from a distribution characterized by the model.

    Returns
    -------
    array_like"""
    import scipy.integrate as integrate
    return (np.log([
        integrate.quad(
            lambda theta: (np.exp(func(Y, f, X+theta * XERR)[i]) * np.exp(-theta*theta/2)/sqrt2pi),
            -np.inf, np.inf)[0] for i, (X, Y, XERR) in enumerate(zip(x, y, xerr))
        ]).sum())

def likelihood_lnprob(params, f, x, y, xerr, func):
    """Calculates the logarithm of the probability that the data fits
    the model given the current parameters.

    Parameters
    ----------
    params: lmfit.Parameters object with satlas.PriorParameters
        Group of parameters for which the fit has to be evaluated.
    f: :class:`.BaseModel`
        Model object containing all the information about the fit;
        will be fitted to the given data.
    x: array_like
        Experimental data for the x-axis.
    y: array_like
        Experimental data for the y-axis.
    xerr: array_like
        Uncertainty values on *x*.
    func: function
        Function calculating the loglikelihood of y_data being drawn from
        a distribution characterized by y_model.

    Note
    ----
    The prior is first evaluated for the parameters. If this is
    not finite, the values are rejected from consideration by
    immediately returning -np.inf."""
    # Handle old-style BaseModel children by using .lnprior().
    try:
        lp = f.get_lnprior_mapping(params)
    except AttributeError:
        lp = f.lnprior()
    f.params = params
    if not np.isfinite(lp):
        return -np.inf
    res = lp + np.sum(likelihood_loglikelihood(f, x, y, xerr, func))
    return res

def likelihood_loglikelihood(f, x, y, xerr, func):
    """Given a parameters object, a Model object, experimental data
    and a loglikelihood function, calculates the loglikelihood for
    all data points.

    Parameters
    ----------
    f: :class:`.BaseModel`
        Model object containing all the information about the fit;
        will be fitted to the given data.
    x: array_like
        Experimental data for the x-axis.
    y: array_like
        Experimental data for the y-axis.
    xerr: array_like
        Experimental data on *x*.
    func: function
        Function calculating the loglikelihood of y_data being drawn from
        a distribution characterized by y_model.

    Returns
    -------
    array_like
        Array containing the loglikelihood for each seperate datapoint."""
    # If a value is given to the uncertainty on the x-values, use the adapted
    # function.
    y = np.hstack(y)
    if xerr is None or np.allclose(0, xerr):
        return_value = func(y, f, x)
    else:
        return_value = likelihood_x_err(f, x, y, xerr, func)
    return return_value

def likelihood_fit(f, x, y, xerr=None, func=llh.poisson_llh, method='nelder-mead', method_kws={}, walking=False, walk_kws={}, verbose=True, hessian=True):
    """Fits the given model to the given data using the Maximum Likelihood Estimation technique.
    The given function is used to calculate the loglikelihood. After the fit, the message
    from the optimizer is printed and returned.

    Parameters
    ----------
    f: :class:`.BaseModel`
        Model to be fitted to the data.
    x: array_like
        Experimental data for the x-axis.
    y: array_like
        Experimental data for the y-axis.

    Other parameters
    ----------------
    xerr: array_like, optional
        Estimated value for the uncertainty on the x-values.
        Set to *None* to ignore this uncertainty. Defaults to *None*.
    func: function, optional
        Used to calculate the loglikelihood that the data is drawn
        from a distribution given a model value. Should accept
        input as (y_data, y_model). Defaults to the Poisson
        loglikelihood.
    method: str, optional
        Selects the algorithm to be used by the minimizer used by LMFIT.
        See LMFIT documentation for possible values. Defaults to 'nelder-mead'.
    method_kws: dict, optional
        Dictionary containing the keywords to be passed to the
        minimizer.
    walking: boolean, optional
        If True, the uncertainty on the parameters is estimated
        by performing a random walk in parameter space and
        evaluating the loglikelihood. Defaults to False.
    walk_kws: dictionary
        Contains the keywords for the :func:`.likelihood_walk`
        function, used if walking is set to True.
    verbose: boolean, optional
        When set to *True*, a tqdm-progressbar in the terminal is maintained.
        Defaults to *True*.
    hessian: boolean, optional
        When set to *True*, the Hessian estimate of the uncertainty will be
        calculated.

    Returns
    -------
    success, message: tuple
        Boolean indicating the success of the optimization and
        the message from the optimizer."""

    def negativeloglikelihood(*args, **kwargs):
        return_val = -likelihood_lnprob(*args, **kwargs)
        return return_val

    def iter_cb(params, iter, resid, *args, **kwargs):
        if verbose:
            progress.update(1)
            progress.set_description('Likelihood fitting in progress (' + str(resid) + ')')
        else:
            pass

    y = np.hstack(y)
    params = copy.deepcopy(f.params)
    # Eliminate the estimated uncertainties
    for p in params:
        params[p].stderr = None
    if verbose:
        progress = tqdm.tqdm(leave=True, desc='Likelihood fitting in progress')

    result = lm.Minimizer(negativeloglikelihood, params, fcn_args=(f, x, y, xerr, func), iter_cb=iter_cb)
    result = result.minimize(method=method, params=params, **method_kws)
    f.params = copy.deepcopy(result.params)
    val = negativeloglikelihood(f.params, f, x, y, xerr, func)
    success = False
    counter = 0
    while not success:
        result = lm.Minimizer(negativeloglikelihood, result.params, fcn_args=(f, x, y, xerr, func), iter_cb=iter_cb)
        result.scalar_minimize(method=method, **method_kws)
        counter += 1
        f.params = copy.deepcopy(result.params)
        new_val = negativeloglikelihood(f.params, f, x, y, xerr, func)
        success = np.isclose(val, new_val)
        val = new_val
        if not success and counter > 10:
            break
    if verbose:
        progress.set_description('Likelihood fitting done')
        progress.close()
    f.ndof_mle = copy.deepcopy(result.nfree)
    f.fit_mle = copy.deepcopy(result.params)
    f.result_mle = result.message
    f.likelihood_mle = negativeloglikelihood(f.params, f, x, y, xerr, func)
    try:
        f.chisqr_mle = np.sum(-2 * likelihood_loglikelihood(f, x, y, xerr, func) + 2 * likelihood_loglikelihood(lambda i: y, x, y, xerr, func))
    except AttributeError:
        f.chisqr_mle = np.nan
    # if np.isnan(f.chisqr_mle):
    #     print('Used loglikelihood does not allow calculation of reduced chisquare for these data points! Does it contain 0 or negative numbers?')
    try:
        f.redchi_mle = f.chisqr_mle / f.ndof_mle
    except:
        f.redchi_mle = f.chisqr_mle / (len(y) - len([p for p in f.params if f.params[p].vary]))

    if hessian:
        if verbose:
            progress = tqdm.tqdm(leave=True, desc='Starting Hessian calculation')
        else:
            progress = None

        assign_hessian_estimate(likelihood_lnprob, f, f.fit_mle, x, y, xerr, func, likelihood=True, progress=progress)
        f.params = copy.deepcopy(f.fit_mle)

    if walking:
        likelihood_walk(f, x, y, xerr=xerr, func=func, **walk_kws)
    return success, result.message

############################
# UNCERTAINTY CALCULATIONS #
############################

def _parameterCostfunction(f, params, func, *args, **kwargs):
    # Creates a costfunction for the given model and arguments/data for the different methods.
    # Is used for calculation the derivative of the cost function for the different parameters.
    likelihood = kwargs.pop('likelihood', False)
    var_names = []
    vars = []
    for key in params.keys():
        if params[key].vary:
            var_names.append(key)
            vars.append(params[key].value)
    if vars == []:
        return
    groupParams = lm.Parameters()
    for key in params.keys():
        groupParams[key] = PriorParameter(key,
                                          value=params[key].value,
                                          vary=params[key].vary,
                                          expr=None,
                                          priormin=params[key].min,
                                          priormax=params[key].max)
    for key in params.keys():
        groupParams[key].expr = params[key].expr

    def listfunc(fvars):
        for val, n in zip(fvars, var_names):
            groupParams[n].value = val
        return func(groupParams, f, *args)
    return listfunc

def assign_hessian_estimate(func, f, params, *args, **kwargs):
    """Calculates the Hessian of the model at the given parameters,
    and associates uncertainty estimates based on the inverted Hessian matrix.
    Note that, for estimation for chisquare methods, the inverted matrix is
    multiplied by 2.

    Parameters
    ----------
    func: function
        Function used as cost function. Use :func:`.chisquare_model` for chisquare estimates,
        and :func:`.likelihood_lnprob` for likelihood estimates.
    f: :class:`.BaseModel`
        Model for which the estimates need to be made.
    params: Parameters
        LMFIT parameters for which the Hessian estimate needs to be made.
    args: args for func
        Arguments for the defined cost function *func*.
    likelihood: boolean
        Set to *True* if a likelihood approach is used.
    progress: progressbar
        TQDM progressbar instance to update.

    Returns
    -------
    None"""
    likelihood = kwargs.pop('likelihood', False)
    progress = kwargs.pop('progress', False)
    if progress is not None:
        progress.set_description('Parsing parameters')
        progress.update(1)

    var_names = []
    vars = []
    for key in params.keys():
        if params[key].vary:
            var_names.append(key)
            vars.append(params[key].value)
    if vars == []:
        if progress is not None:
            progress.set_description('No parameters to vary')
            progress.update(1)
            progress.close()
        return

    if progress is not None:
        progress.set_description('Creating Hessian function')
        progress.update(1)
    Hfun = nd.Hessian(_parameterCostfunction(f, params, func, *args, **kwargs))
    if progress is not None:
        progress.set_description('Calculating Hessian matrix')
        progress.update(1)
    hess_vals = Hfun(vars)
    if progress is not None:
        progress.set_description('Inverting matrix')
        progress.update(1)
    hess_vals = np.linalg.inv(hess_vals)
    if likelihood:
        hess_vals = -hess_vals
        multiplier = 1
    else:
        multiplier = 2
    if progress is not None:
        progress.set_description('Assigning uncertainties')
        progress.update(1)
    for name, hess in zip(var_names, np.diag(multiplier*hess_vals)):
        params[name].stderr = np.sqrt(hess)

    if progress is not None:
        progress.set_description('Assigning correlations')
        progress.update(1)
    for i, name in enumerate(var_names):
        params[name].correl = {}
        for j, name2 in enumerate(var_names):
            if name != name2:
                params[name].correl[name2] = hess_vals[i, j] / np.sqrt(hess_vals[i, i]*hess_vals[j, j])
    if progress is not None:
        progress.set_description('Finished Hessian calculation')
        progress.update(1)
        progress.close()

def create_band(f, x, x_data, y_data, yerr, xerr=None, method='chisquare', func_chi=None, func_llh=llh.poisson_llh, kind='prediction'):
    r"""Calculates prediction or confidence bounds at the 1 :math:`\sigma` level.
    The method used is based on the Delta Method: at the requested prediction points *x*, the bound is calculated as

    .. math::
        \sqrt{G'(\beta, x)^T H^{-1}(\beta) G'(\beta, x)}

    with G the cost function, H the Hessian matrix and :math:`\beta` the vector of parameters.
    The resulting bound needs to be subtracted and added to the value given by the model to get the confidence interval.

    For a prediction interval, the value before taking the square root is increased by 1


    Parameters
    ----------
    f: :class:`.BaseModel`
        Model for which the bound needs to be calculated.
    x: array_like
        Selection of values for which a prediction needs to be made.
    x_data: array_like
        Experimental data for the x-axis.
    y_data: array_like
        Experimental data for the y-axis.
    yerr: array_like
        Experimental uncertainty for the y-axis.
    xerr: array_like
        Experimental uncertainty for the x-axis. Defaults to *None*.
    method: {'mle', 'chisquare'}
        Selected method for which the cost function is selected.
    func_chi: function, optional
        Is passed on to the chisquare methods in order to calculate the
        experimental uncertainty from the modelvalue. Defaults to *None*,
        which uses *yerr*.
    func_llh: function
        Is passed on to the likelihood fitting method to define the
        likelihood function. Defaults to :func:`satlas.loglikelihood.poisson_llh`.
    kind: {'prediction', 'confidence'}
        Selects which type of bound is calculated.

    Returns
    -------
    bound: array_like
        Array describing the deviation from the model value as can be expected
        for the selected parameters at the 1:math:`\sigma` level."""
    method_mapping = {'mle': likelihood_lnprob,
                      'chisquare': lambda *args: (chisquare_model(*args)**2).sum()}
    if method == 'chisquare':
        args = x_data, np.hstack(y_data), np.hstack(yerr), xerr, func_chi
    else:
        args = x_data, y_data, xerr, func_llh, False
    func = method_mapping.pop(method)
    var_names = []
    vars = []
    for key in f.params.keys():
        if f.params[key].vary:
            var_names.append(key)
            vars.append(f.params[key].value)

    backup = copy.deepcopy(f.params)
    Hfun = nd.Hessian(_parameterCostfunction(f, f.params, func, *args, likelihood=method.lower()=='mle'))
    if method.lower()=='mle':
        hess_vals = -np.linalg.inv(Hfun(vars))
    else:
        hess_vals = np.linalg.inv(Hfun(vars))*2
    groupParams = lm.Parameters()
    for key in f.params.keys():
        groupParams[key] = PriorParameter(key,
                                          value=f.params[key].value,
                                          vary=f.params[key].vary,
                                          expr=f.params[key].expr,
                                          priormin=f.params[key].min,
                                          priormax=f.params[key].max)
    def listfunc(fvars):
        for val, n in zip(fvars, var_names):
            groupParams[n].value = val
        f.params = groupParams
        return f(x)
    jacob = nd.Jacobian(listfunc)
    result = np.zeros(len(x))
    for i, row in enumerate(jacob(vars)):
        result[i] = np.dot(row.T, np.dot(hess_vals, row))
    f.params = backup
    if kind.lower()=='prediction':
        return (result+1)**0.5
    else:
        return result**0.5

params_map = {'mle': 'fit_mle', 'chisquare_spectroscopic': 'chisq_res_par', 'chisquare': 'chisq_res_par'}
fit_mapping = {'mle': likelihood_fit, 'chisquare_spectroscopic': chisquare_spectroscopic_fit, 'chisquare': chisquare_fit}
attr_mapping = {'mle': 'likelihood_mle', 'chisquare_spectroscopic': 'chisqr_chi', 'chisquare': 'chisqr_chi'}

def calculate_updated_statistic(value, params_name, f, x, y, method='chisquare', func_args=tuple(), func_kwargs={}, pbar=None, orig_stat=0):
    params = copy.deepcopy(f.params)
    func = fit_mapping[method.lower()]
    attr = attr_mapping[method.lower()]

    try:
        for v, n in zip(value, params_name):
            params[n].value = v
            params[n].vary = False
    except:
        params[params_name].value = value
        params[params_name].vary = False

    f.params = params
    success = False
    counter = 0
    while not success:
        try:
            success, message = func(f, x, y, *func_args, **func_kwargs)
        except ValueError:
            f.params.pretty_print()
            raise
        counter += 1
        if counter > 10:
            return np.nan
    return_value = getattr(f, attr) - orig_stat
    try:
        try:
            params_name = ' '.join(params_name)
        except:
            pass
        pbar.set_description(params_name + ' (' + str(value, return_value) + ')')
        pbar.update(1)
    except:
        pass
    return return_value

def _find_boundary(step, param_name, bound, f, x, y, function_kwargs={'method': 'chisquare_spectroscopic'}, verbose=True):
    method = function_kwargs['method']
    attr = attr_mapping[method.lower()]
    orig_stat = getattr(f, attr)
    value = f.params[param_name].value
    search_value = value
    if step < 0:
        direction = 'left'
        boundary = f.params[param_name].min
        boundary_test = lambda val: val < boundary
    else:
        direction = 'right'
        boundary = f.params[param_name].max
        boundary_test = lambda val: val > boundary
    if verbose:
        pbar = tqdm.tqdm(leave=True, desc=param_name + ' (searching ' + direction + ')', miniters=1)
    else:
        pbar = None
    backup = copy.deepcopy(f.params)
    backup_fit = params_map[method.lower()]
    function_kwargs['pbar'] = pbar
    function_kwargs['orig_stat'] = orig_stat
    while True:
        search_value += step
        if boundary_test(search_value):
            try:
                pbar.set_description(desc=param_name + ' (' + direction + ' limit reached)')
                pbar.update(1)
            except:
                pass
            result = boundary
            success = True
            break
        new_value = calculate_updated_statistic(search_value, param_name, f, x, y, **function_kwargs)
        try:
            pbar.set_description(desc=param_name + ' (searching ' + direction + ':  {:.3g}, change of {:.3f}, at {:.3f}%)'.format(search_value, new_value, new_value/bound*100))
            pbar.update(1)
        except:
            pass
        if new_value > bound:
            try:
                pbar.set_description(desc=param_name + ' (finding root, between {:.3g} and {:.3g})'.format(value, search_value))
                pbar.update(1)
            except:
                pass
            result, output = optimize.bisect(lambda v: calculate_updated_statistic(v, param_name, f, x, y, **function_kwargs) - bound,
                                             search_value - step, search_value,
                                             full_output=True)
            try:
                pbar.set_description(desc=param_name + ' (root found: {:.3g})'.format(result))
                pbar.update(1)
            except:
                pass
            success = output.converged
            break
    result_value = calculate_updated_statistic(result, param_name, f, x, y, **function_kwargs)
    try:
        pbar.set_description(desc=param_name + ' (root found: {:.3g}, change of {:.3f})'.format(result, result_value))
        pbar.update(1)
        pbar.close()
    except:
        pass
    f.params = copy.deepcopy(backup)
    setattr(f, attr, orig_stat)
    return result, success

def _get_state(f, method='mle'):
    if method.lower() == 'mle':
        return (copy.deepcopy(f.fit_mle), f.result_mle, f.likelihood_mle, f.chisqr_mle, f.redchi_mle)
    else:
        return (copy.deepcopy(f.chisq_res_par), f.chisqr_chi, f.ndof_chi, f.redchi_chi)

def _set_state(f, state, method='mle'):
    if method.lower() == 'mle':
        f.fit_mle = copy.deepcopy(state[0])
        f.params = copy.deepcopy(state[0])
        f.result_mle, f.likelihood_mle, f.chisqr_mle, f.redchi_mle = state[1:]
    else:
        f.chisq_res_par = copy.deepcopy(state[0])
        f.params = copy.deepcopy(state[0])
        f.chisqr_chi, f.ndof_chi, f.redchi_chi = state[1:]

def calculate_analytical_uncertainty(f, x, y, method='chisquare_spectroscopic', filter=None, fit_args=tuple(), fit_kws={}):
    """Calculates the analytical errors on the parameters, by changing the value for
    a parameter and finding the point where the chisquare for the refitted parameters
    is one greater. For MLE, an increase of 0.5 is sought. The corresponding series
    of parameters of the model is adjusted with the values found here.

    Parameters
    ----------
    f: :class:`.BaseModel`
        Instance of a model which is to be fitted.
    x: array_like
        Experimental data for the x-axis.
    y: array_like
        Experimental data for the y-axis.

    Other parameters
    ----------------
    method: {'chisquare_spectroscopic', 'chisquare', 'mle'}
        Select for which method the analytical uncertainty has to be calculated.
        Defaults to 'chisquare_spectroscopic'.
    filter: list of strings, optional
        Select only a subset of the variable parameters to calculate the uncertainty for.
        Defaults to *None* (all parameters).
    fit_kws: dictionary, optional
        Dictionary of keywords to be passed on to the selected fitting routine.

    Note
    ----
    The function edits the parameters of the given instance. Furthermore,
    it only searches for the uncertainty in the neighbourhood of the starting
    point, which is taken to be the values of the parameters as given in
    the instance. This does not do a full exploration, so the results might be
    from a local minimum!"""

    # Save the original goodness-of-fit and parameters for later use
    mapping = {'chisquare_spectroscopic': (chisquare_spectroscopic_fit, 'chisqr', 'chisq_res_par'),
               'chisquare': (chisquare_fit, 'chisqr', 'chisq_res_par'),
               'mle': (likelihood_fit, 'likelihood_mle', 'fit_mle')}
    func, attr, save_attr = mapping.pop(method.lower(), (chisquare_spectroscopic_fit, 'chisqr', 'chisq_res_par'))
    fit_kws['verbose'] = False
    fit_kws['hessian'] = False

    func(f, x, y, *fit_args, **fit_kws)

    state = _get_state(f, method=method.lower())
    orig_params = state[0]
    f.params = copy.deepcopy(state[0])

    ranges = {}

    # Select all variable parameters, generate the figure
    param_names = []
    no_params = 0
    for p in orig_params:
        if orig_params[p].vary and (filter is None or any([f in p for f in filter])):
            no_params += 1
            param_names.append(p)

    params = copy.deepcopy(f.params)
    chifunc = lambda x: chi2.cdf(x, 1) - 0.682689492 # Calculate 1 sigma boundary
    bound = optimize.root(chifunc, 1).x[0] * 0.5 if method.lower() == 'mle' else optimize.root(chifunc, 1).x[0]
    for i in range(no_params):
        # _set_state(f, state, method=method.lower())
        ranges[param_names[i]] = {}
        # Select starting point to determine error widths.
        value = orig_params[param_names[i]].value
        stderr = orig_params[param_names[i]].stderr
        stderr = stderr if stderr is not None else 0.01 * np.abs(value)
        stderr = stderr if stderr != 0 else 0.01 * np.abs(value)

        function_kws = {'method': method.lower(), 'func_args': fit_args, 'func_kwargs': fit_kws}

        result_left, success_left = _find_boundary(stderr, param_names[i], bound, f, x, y, function_kwargs=function_kws)
        result_right, success_right = _find_boundary(-stderr, param_names[i], bound, f, x, y, function_kwargs=function_kws)
        success = success_left * success_right
        ranges[param_names[i]]['left'] = result_left
        ranges[param_names[i]]['right'] = result_right

        if not success:
            print("Warning: boundary calculation did not fully succeed for " + param_names[i])
        right = np.abs(ranges[param_names[i]]['right'] - value)
        left = np.abs(ranges[param_names[i]]['left'] - value)
        ranges[param_names[i]]['uncertainty'] = max(right, left)
        ranges[param_names[i]]['value'] = orig_params[param_names[i]].value

    # First, clear all uncertainty estimates
    for p in orig_params:
        orig_params[p].stderr = None
    for param_name in ranges.keys():
        orig_params[param_name].stderr = ranges[param_name]['uncertainty']
        orig_params[param_name].value = ranges[param_name]['value']
    state = list(state)
    state[0] = copy.deepcopy(orig_params)
    state = tuple(state)
    _set_state(f, state, method=method.lower())

def likelihood_walk(f, x, y, xerr=None, func=llh.poisson_llh, nsteps=2000, walkers=20, filename=None):
    """Calculates the uncertainty on MLE-optimized parameter values
    by performing a random walk through parameter space and comparing
    the resulting loglikelihood values. For more information,
    see the emcee package. The data from the random walk is saved in a
    file, as defined with the *filename*.

    Parameters
    ----------
    f: :class:`.BaseModel`
        Model to be fitted to the data.
    x: array_like
        Experimental data for the x-axis.
    y: array_like
        Experimental data for the y-axis.

    Other parameters
    ----------------
    func: function, optional
        Used to calculate the loglikelihood that the data is drawn
        from a distribution given a model value. Should accept
        input as (y_data, y_model). Defaults to the Poisson
        loglikelihood.
    nsteps: integer, optional
        Determines how many steps each walker should take.
        Defaults to 2000 steps.
    walkers: integer, optional
        Sets the number of walkers to be used for the random walk.
        The number of walkers should never be less than twice the
        number of parameters. For more information on this, see
        the emcee documentation. Defaults to 20 walkers.
    filename: string, optional
        Filename where the random walk has to be saved. If *None*,
        the current time in seconds since January 1970 is used.

    Note
    ----
    The parameters associated with the MLE fit are not updated
    with the uncertainty as estimated by this method."""

    params = f.params
    var_names = []
    vars = []
    for key in params.keys():
        if params[key].vary and params[key].expr is None:
            var_names.append(key)
            vars.append(params[key].value)
    ndim = len(vars)
    pos = mcmc.utils.sample_ball(vars, [1e-4] * len(vars), size=walkers)
    for i in range(pos.shape[1]):
        pos[:, i] = np.where(pos[:, i] < params[var_names[i]].min, params[var_names[i]].min+(1E-5), pos[:, i])
        pos[:, i] = np.where(pos[:, i] > params[var_names[i]].max, params[var_names[i]].max-(1E-5), pos[:, i])

    def lnprobList(fvars, groupParams, f, x, y, xerr, func):
        for val, n in zip(fvars, var_names):
            groupParams[n].value = val
        return likelihood_lnprob(groupParams, f, x, y, xerr, func)

    p = f.params.copy()
    groupParams = f.params.copy()
    for key in params.keys():
        groupParams[key] = PriorParameter(key,
                                          value=params[key].value,
                                          vary=params[key].vary,
                                          expr=params[key].expr,
                                          priormin=params[key].min,
                                          priormax=params[key].max)
    sampler = mcmc.EnsembleSampler(walkers, ndim, lnprobList,
                                   args=(groupParams, f, x, y, xerr, func))

    if filename is None:
        import time
        filename = '{}.h5'.format(time.time())
    else:
        filename = '.'.join(filename.split('.')[:-1]) + '.h5'

    if not os.path.isfile(filename):
        with h5py.File(filename, 'w') as store:
            dset = store.create_dataset('data', (nsteps * walkers, ndim), dtype='float', chunks=True, compression='gzip', maxshape=(None, ndim))
            dset.attrs['format'] = np.array([f.encode('utf-8') for f in var_names])
            initial_length = 0
    else:
        with h5py.File(filename, 'a') as store:
            dset = store['data']
            pos = dset[-walkers:, :]
            initial_length = dset.shape[0]

    with tqdm.tqdm(total=nsteps, desc='Walk', leave=True) as pbar:
        for i, result in enumerate(sampler.sample(pos, iterations=nsteps, storechain=False)):
            with h5py.File(filename, 'a') as store:
                dset = store['data']
                dset.resize((initial_length+(i+1)*walkers, ndim))
                dset[initial_length+i*walkers:initial_length+(i+1)*walkers,:] = result[0]
                pbar.update(1)

    f.fit_mle = copy.deepcopy(params)
    f.params = copy.deepcopy(params)

def process_walk(model, filename, selection=(0, 100)):
    r"""Given a model and H5 file with the results of a random walk,
    the parameters varied in the walk are set to the 50% percentile,
    and the uncertainty is set to either the difference between the
    50% and 16% percentile, or between the 84% and 50%, whichever is
    the largest.

    Parameters
    ----------
    model: :class:`.BaseModel`
        Object which has a params attribute.
    filename: str
        Filename of the corresponding H5 file.

    Other parameters
    ----------------
    selection: tuple
        Sets the lower and upper boundary of the percentage of the random walk
        to take into account. Can be used to eliminate burn-in.
        Defaults to (0, 100)."""
    p = model.params.copy()
    with h5py.File(filename, 'r') as store:
        columns = store['data'].attrs['format']
        columns = [f.decode('utf-8') for f in columns]
        with tqdm.tqdm(total=len(columns)+(len(columns)**2-len(columns))/2, leave=True) as pbar:
            dataset_length = store['data'].shape[0]
            first, last = int(np.floor(dataset_length/100*selection[0])), int(np.ceil(dataset_length/100*selection[1]))
            for i, val in enumerate(columns):
                pbar.set_description(val)
                i = columns.index(val)
                x = store['data'][first:last, i]

                q = [16.0, 50.0, 84.0]
                q16, q50, q84 = np.percentile(x, q)

                value, std_dev = q50, max(q50-q16, q84-q50)
                p[val].value = value
                p[val].stderr = std_dev
                pbar.update(1)
    model.params = p.copy()
    model.fit_mle = p.copy()
    model.chisqr_mle = np.nan
    model.redchi_mle = np.nan
